"""Module used to create GUIs and handle backend processing for user file input and output.
"""

from tkinter import BOTH, filedialog
from tkinter import ttk
import tkinter.simpledialog as sd
import tkinter as tk
from time import sleep
import sys
import os
import importlib
import inspect
import json
from typing import Callable, Tuple
import numpy as np
import time

from embrs.utilities.fire_util import FuelConstants
from embrs.base_classes.control_base import ControlClass

class FileSelectBase:
    """Base class for creating tkinter file and folder selector interfaces

    :param title: tile to be displayed at the top of the window
    :type title: str
    """
    def __init__(self, title: str):
        """Constructor method that creates a tk root and sets the title
        """
        self.root = tk.Tk()
        self.root.title(title)
        self.result = None

    def create_frame(self, tar: tk.Frame) -> tk.Frame:
        """Create a new tkinter frame within 'tar'

        :param tar: target tk.Frame that the new frame will be created within
        :type tar: tk.Frame
        :return: tk.Frame within 'tar' that can be used to add tkinter elements to
        :rtype: tk.Frame
        """
        frame = tk.Frame(tar)
        frame.pack(fill=BOTH, padx=5, pady=5)
        return frame

    def create_entry_with_label_and_button(self, frame: tk.Frame, text: str, text_var: any,
                                           button_text: str,button_command: Callable
                                           ) -> Tuple[tk.Label,tk.Entry, tk.Button, tk.Frame]:
        """Creates a tk.Frame with a label, entry, and button.

        :param frame: Root frame where the new frame should be located
        :type frame: tk.Frame
        :param text: Text that will displayed on the label
        :type text: str
        :param text_var: Variable where the data entered in the entry will be stored
        :type text_var: any
        :param button_text: Button text to let user know what the button does
        :type button_text: str
        :param button_command: Callable function that should be triggered when the button is
                               pressed
        :type button_command: Callable
        :return: The resulting tk.Label, tk.Entry, tk.Button, and tk.Frame objects
        :rtype: Tuple[tk.Label, tk.Entry, tk.Button, tk.Frame]
        """
        new_frame = self.create_frame(frame)

        label = tk.Label(new_frame, text=text, anchor="w")
        label.grid(row=0, column=0)

        entry = tk.Entry(new_frame, textvariable=text_var, width=50)
        entry.grid(row=0, column=1, sticky="w")

        button = tk.Button(new_frame, text=button_text, command=button_command)
        button.grid(row=0, column=2)

        return label, entry, button, new_frame

    def create_file_selector(self, frame: tk.Frame, text: str, text_var: any, file_type:list=None
                            ) -> Tuple[tk.Label, tk.Entry, tk.Button, tk.Frame]:
        """Creates an entry, label, and button that operate as a way to select a specific file
        path from user's device

        :param frame: Root frame where the new frame should be located
        :type frame: tk.Frame
        :param text: Text that will displayed on the label
        :type text: str
        :param text_var: Variable where the path entered in the entry will be stored
        :type text_var: any
        :param file_type: List of acceptable file types to be selected in the form
                          [(description, file extension)], if None all files will be selectable,
                          defaults to None
        :type file_type: list, optional
        :return: The resulting tk.Label, tk.Entry, tk.Button, and tk.Frame objects
        :rtype: Tuple[tk.Label, tk.Entry, tk.Button, tk.Frame]
        """
        label, entry, button, frame = self.create_entry_with_label_and_button(frame, text,
                                                text_var, "Browse",
                                                lambda: self.select_file(text_var, file_type))

        return label, entry, button, frame
    def create_folder_selector(self, frame: tk.Frame, text: str, text_var: any
                              ) -> Tuple[tk.Label, tk.Entry, tk.Button, tk.Frame]:
        """Creates an entry, label, and button that operate as a way to select a specific folder
        path from user's device

        :param frame: Root frame where the new frame should be located
        :type frame: tk.Frame
        :param text: Text that will displayed on the label
        :type text: str
        :param text_var: Variable where the path entered in the entry will be stored
        :type text_var: any
        :return: The resulting tk.Label, tk.Entry, tk.Button, and tk.Frame objects
        :rtype: Tuple[tk.Label, tk.Entry, tk.Button, tk.Frame]
        """
        label, entry, button, frame = self.create_entry_with_label_and_button(frame, text,
                                                    text_var, "Browse",
                                                    lambda: self.select_folder(text_var))

        return label, entry, button, frame

    def create_spinbox_with_two_labels(self, frame: tk.Frame, left_label:str, max_val:float,
                                       var: any, right_label:str):
        """Creates a frame containing a spinbox with one or two labels

        :param frame: Root frame where the new frame should be located 
        :type frame: tk.Frame
        :param left_label: Label to be displayed to the left of the spinbox
        :type left_label: str
        :param max_val: Max allowable value in the spinbox
        :type max_val: float
        :param var: Variable where the result of the spinbox entry should be stored
        :type var: any
        :param right_label: Label to be displayed to the right of the spinbox
        :type right_label: str
        """
        new_frame = self.create_frame(frame)
        tk.Label(new_frame, text=left_label).grid(row=0, column=0)
        tk.Spinbox(new_frame, from_=1, to=max_val, textvariable=var,
                   width=5).grid(row=0, column=1, sticky='w')

        if right_label is not None:
            tk.Label(new_frame, text=right_label).grid(row=0, column=2)

    def select_file(self, text_var: any, file_type: list):
        """Function that opens a filedialog window and prompts user to select a file from their
        device

        :param text_var: Variable where the path selected will be stored
        :type text_var: any
        :param file_type: List of acceptable file types to be selected in the form
                          [(description, file extension)], if None all files will be selectable
        :type file_type: list
        """
        filepath = filedialog.askopenfilename(filetypes=file_type)

        if filepath != "":
            text_var.set(filepath)
            self.validate_fields()

    def select_folder(self, text_var: any):
        """Function that opens a filedialog window and prompts user to select a folder from their
        device

        :param text_var: Variable where the path selected will be stored
        :type text_var: any
        """
        folderpath = filedialog.askdirectory()

        if folderpath != "":
            text_var.set(folderpath)
            self.validate_fields()

    def run(self):
        """Runs the FileSelectBase instance and returns the result variable

        :return: Results of the run, typically a dictionary containing input data from user
        :rtype: any
        """
        self.root.mainloop()
        return self.result

    def validate_fields(self):
        """Function to validate inputs that all instances should implement

        :raises NotImplementedError: not implemented here
        """
        raise NotImplementedError

    def submit(self):
        """Function to link to submit button to signal the end of the input process that all
        instances should implement

        :raises NotImplementedError: not implemented here
        """
        raise NotImplementedError

class MapGenFileSelector(FileSelectBase):
    """Class used to prompt user for map generation files
    """
    def __init__(self):
        """Constructor method, populates tk window with all necessary elements and initializes
        necessary variables
        """

        super().__init__("Select Input Data Files and Map Destination Folder")

        # Define variables
        self.uniform_fuel = tk.BooleanVar()
        self.uniform_fuel.trace("w", self.uniform_options_toggled)
        self.fuel_selection = tk.StringVar()
        self.fuel_selection.set("Short grass")
        self.fuel_selection.trace("w", self.fuel_selection_changed)
        self.fuel_selection_val = 1
        self.fuel_map_filename = tk.StringVar()
        self.uniform_elev = tk.BooleanVar()
        self.uniform_elev.trace("w", self.uniform_options_toggled)
        self.topography_map_filename = tk.StringVar()
        self.output_map_folder = tk.StringVar()
        self.import_roads = tk.BooleanVar()
        self.import_roads.set(False)
        self.sim_width = tk.IntVar()
        self.sim_width.set(1000)
        self.sim_height = tk.IntVar()
        self.sim_height.set(1000)

        frame = self.create_frame(self.root)

        # Create field to select save destination
        self.create_folder_selector(frame, "Save map to:   ", self.output_map_folder)

        # Create frame for fuel map selection
        _, _, self.fuel_button, self.fuel_frame = self.create_file_selector(frame, "Fuel Map:         ",
                                                  self.fuel_map_filename,
                                                  [("Tagged Image File Format","*.tif"),
                                                  ("Tagged Image File Format","*.tiff")])

        # Create field for uniform fuel option
        tk.Checkbutton(self.fuel_frame, text='Uniform Fuel',
                       variable=self.uniform_fuel).grid(row=0, column=3)

        uniform_fuel_frame = tk.Frame(frame)
        uniform_fuel_frame.pack(padx=10,pady=5)
        tk.Label(uniform_fuel_frame, text="Uniform Fuel Type:",
                 anchor="center").grid(row=0, column=0)

        tk.OptionMenu(uniform_fuel_frame, self.fuel_selection,
                      *FuelConstants.fuel_names.values()).grid(row=0, column=1)

        # Create frame for topography map selection
        _, _, self.elev_button, self.elev_frame = self.create_file_selector(frame,
                                                  "Elevation Map: ", self.topography_map_filename,
                    [("Tagged Image File Format","*.tif"), ("Tagged Image File Format","*.tiff")])

        # Create field for uniform elev option
        tk.Checkbutton(self.elev_frame, text="Uniform Elevation",
                       variable=self.uniform_elev).grid(row=0, column=3)


        # Create frame for importing roads
        import_road_frame = tk.Frame(frame)
        import_road_frame.pack(padx=10,pady=5)
        self.import_roads_button = tk.Checkbutton(import_road_frame,
                                   text='Import Roads from OpenStreetMap',
                                   variable = self.import_roads, anchor="center")

        self.import_roads_button.grid(row=0, column=0)

        # Create frame for specifying sim size
        sim_size_frame = tk.Frame(frame)
        sim_size_frame.pack(padx=10,pady=5)
        tk.Label(sim_size_frame, text="width (m)", anchor="center").grid(row=0, column=0)
        self.width_entry = tk.Entry(sim_size_frame, textvariable=self.sim_width, width=20,
                                                                         state='disabled')

        self.width_entry.grid(row=0, column=1)
        tk.Label(sim_size_frame, text="height (m)", anchor="center").grid(row=0, column=2)
        self.height_entry = tk.Entry(sim_size_frame, textvariable=self.sim_height, width=20,
                                     state='disabled')

        self.height_entry.grid(row=0, column=3)

        # Create a submit button
        self.submit_button = tk.Button(frame, text="Submit", command=self.submit, state='disabled')
        self.submit_button.pack(pady=10)

    def uniform_options_toggled(self, *args):
        """Callback function that handles uniform fuel and uniform elevation options,
        turns on/off functionality based on these selections
        """
        if self.uniform_fuel.get() and self.uniform_elev.get():
            self.fuel_button.configure(state='disabled')
            self.elev_button.configure(state='disabled')
            self.height_entry.config(state='normal')
            self.width_entry.config(state='normal')
            self.import_roads_button.config(state='disabled')
            self.import_roads.set(False)

        elif self.uniform_fuel.get() and not self.uniform_elev.get():
            self.fuel_button.configure(state='disabled')
            self.elev_button.configure(state='active')
            self.height_entry.config(state='disabled')
            self.width_entry.config(state='disabled')
            self.import_roads_button.config(state='active')

        elif not self.uniform_fuel.get() and self.uniform_elev.get():
            self.fuel_button.configure(state='active')
            self.elev_button.configure(state='disabled')
            self.height_entry.config(state='disabled')
            self.width_entry.config(state='disabled')
            self.import_roads_button.config(state='active')

        else:
            self.fuel_button.configure(state='active')
            self.elev_button.configure(state='active')
            self.width_entry.config(state='disabled')
            self.height_entry.config(state='disabled')
            self.import_roads_button.config(state='active')

        self.validate_fields()

    def fuel_selection_changed(self, *args):
        """Callback function to handle the uniform fuel type value being changed
        """
        self.fuel_selection_val = FuelConstants.fuel_type_reverse_lookup[self.fuel_selection.get()]


    def validate_fields(self):
        """Function used to validate the inputs, primarily responsible for activating/disabling
        the submit button based on if all necessary input has been provided.
        """
        # Check that all fields are filled before enabling submit button
        if all([(self.fuel_map_filename.get() or self.uniform_fuel.get()),
                (self.topography_map_filename.get() or self.uniform_elev.get()),
                 self.output_map_folder.get()]):

            self.submit_button.config(state='normal')

        else:
            self.submit_button.config(state='disabled')

    def submit(self):
        """Callback when the submit button is pressed. Stores all the relevant data in the result
        variable so it can be retrieved
        """
        if self.import_roads.get():
            # check that metadata.xml exists
            elev_path = os.path.dirname(self.topography_map_filename.get())
            fuel_path = os.path.dirname(self.fuel_map_filename.get())

            if os.path.exists(elev_path + "/metadata.xml"):
                metadata_path = elev_path + "/metadata.xml"

            elif os.path.exists(fuel_path + "/metadata.xml"):
                metadata_path = fuel_path + "/metadata.xml"

            else:
                self.import_roads.set(False)
                metadata_path = ""
                window = tk.Tk()
                window.withdraw()
                tk.messagebox.showwarning("Error",
                "Cannot locate 'metadata.xml' file, roads will not be imported.")

                window.destroy()
        else:
            metadata_path = ""

        self.result = {}

        self.result["Output Map Folder"] = self.output_map_folder.get()
        self.result["Metadata Path"] = metadata_path
        self.result["Import Roads"] = self.import_roads.get()

        if self.uniform_fuel.get():
            self.result["Uniform Fuel"] = True
            self.result["Fuel Type"] = self.fuel_selection_val
            self.result["Fuel Map Path"] = ""

        else:
            self.result["Uniform Fuel"] = False
            self.result["Fuel Map Path"] = self.fuel_map_filename.get()

        if self.uniform_elev.get():
            self.result["Uniform Elev"] = True
            self.result["Topography Map Path"] = ""
        else:
            self.result["Uniform Elev"] = False
            self.result["Topography Map Path"] = self.topography_map_filename.get()

        if self.uniform_elev.get() and self.uniform_fuel.get():
            self.result["width m"] = self.sim_width.get()
            self.result["height m"] = self.sim_height.get()

        self.root.withdraw()
        self.root.quit()

class SimFolderSelector(FileSelectBase):
    """Class used to prompt user for inputs to set up a sim
    """
    def __init__(self, submit_callback:Callable):
        """Constructor method, populates tk window with all necessary elements and initializes
        necessary variables
        """
        super().__init__("EMBRS")

        self.submit_callback = submit_callback

        # Define variables
        self.map_folder = tk.StringVar()
        self.log_folder = tk.StringVar()
        self.wind_forecast = tk.StringVar()
        self.time_step = tk.IntVar()
        self.cell_size = tk.IntVar()
        self.sim_time = tk.DoubleVar()
        self.time_unit = tk.StringVar()
        self.num_runs = tk.IntVar()
        self.user_path = tk.StringVar()
        self.user_class_name = tk.StringVar()
        self.viz_on = tk.BooleanVar()
        self.zero_wind = tk.BooleanVar()

        # Define some useful variables
        self.prev_t_unit = "hours"
        self.max_duration = np.inf
        self.class_options = []
        self.user_module = None

        # Set some initial values
        self.time_step.set(5)
        self.cell_size.set(10)
        self.sim_time.set(1.0)
        self.num_runs.set(1)
        self.viz_on.set(False)
        self.time_unit.set("hours")
        self.zero_wind.set(False)

        frame = self.create_frame(self.root)

        # Create frame for map folder selection
        self.create_folder_selector(frame, "Map folder:    ", self.map_folder)

        # Create frame for log folder selection
        self.create_folder_selector(frame, "Log folder:     ", self.log_folder)

        # Create frame for wind file selection
        _, self.wind_entry, self.wind_button, self.wind_frame = self.create_file_selector(frame, "Wind forecast: ", self.wind_forecast, [("JavaScript Object Notation","*.JSON")])

        tk.Checkbutton(self.wind_frame, text='No Wind',
                       variable=self.zero_wind).grid(row=0, column=3)

        # Create frame for time step selection
        self.create_spinbox_with_two_labels(frame, "Time step:     ", 100, self.time_step, "seconds")

        # Create frame for cell size selection
        self.create_spinbox_with_two_labels(frame, "Cell size:       ", 100, self.cell_size, "meters")

        # Create frame for sim time selection
        sim_time_frame = self.create_frame(frame)
        tk.Label(sim_time_frame, text="Duration:       ").grid(row=2, column=0)

        self.duration_spinbox = tk.Spinbox(sim_time_frame, from_=1, to=self.max_duration,
                                           textvariable=self.sim_time, width=5, format='%.2f', increment = 0.5)

        self.duration_spinbox.grid(row=2, column=1, sticky='w')
        tk.OptionMenu(sim_time_frame, self.time_unit,
                      "seconds", "minutes", "hours").grid(row=2, column=2)

        # Create frame for num runs selection
        self.create_spinbox_with_two_labels(frame, "Iterations:      ", np.inf,

                                            self.num_runs, None)

        # Create frame for user module selection
        self.create_file_selector(frame, "User module:", self.user_path, [("python file","*.py")])

        # Create frame for user class selection
        class_name_frame = tk.Frame(frame)
        class_name_frame.pack(padx=10,pady=5)
        tk.Label(class_name_frame, text = "User class name:").grid(row=0, column=0)
        self.class_opt_menu = tk.OptionMenu(class_name_frame, self.user_class_name,
                                            self.class_options)
        self.class_opt_menu.grid(row=0, column=1)

        # Create frame for viz selection
        viz_frame = tk.Frame(frame)
        viz_frame.pack(padx=10, pady=5)
        tk.Checkbutton(viz_frame, text='Visualize in Real-time',
                       variable = self.viz_on).grid(row=0, column=0)

        # Create a submit button
        self.submit_button = tk.Button(frame, text="Submit", command=self.submit, state='disabled')
        self.submit_button.pack(pady=10)

        # Define variable callbacks
        self.map_folder.trace("w", self.map_folder_changed)
        self.log_folder.trace("w", self.log_folder_changed)
        self.wind_forecast.trace("w", self.wind_forecast_changed)
        self.sim_time.trace("w", self.duration_changed)
        self.time_unit.trace("w", self.time_unit_changed)
        self.user_path.trace("w", self.user_path_changed)
        self.user_class_name.trace("w", self.class_changed)
        self.zero_wind.trace("w", self.zero_wind_toggled)

    def duration_changed(self, *args):
        """Callback function that handles the sim duration changing, prevents values greater than 
        the max being input
        """
        try:
            self.sim_time.get()
        except tk.TclError:
            return

        if self.sim_time.get() > self.max_duration:
            self.sim_time.set(self.max_duration)

    def map_folder_changed(self, *args):
        """Callback function that handles the map folder selection being changed, sets the max
        duration allowed based on the length of the wind forecast in the map
        """
        # open json
        folderpath = self.map_folder.get()
        foldername = os.path.basename(folderpath)
        map_filepath = folderpath + "/" + foldername + ".json"

        if os.path.exists(map_filepath):
            pass

        elif folderpath == "":
            pass

        else:
            self.map_folder.set("")
            window = tk.Tk()
            window.withdraw()
            tk.messagebox.showwarning("Error", "Selected folder does not contain a valid map!")
            window.destroy()

    def log_folder_changed(self, *args):
        """Callback function that handles the log folder selection being changed, warns the user
        if the log folder is already populated with logs so they are aware that those logs will
        be overwritten
        """
        folderpath = self.log_folder.get()

        if os.path.exists(folderpath + '/init_fire_state.pkl') or os.path.exists(folderpath + '/run_0'):
            window = tk.Tk()
            window.withdraw()
            msg = """Warning: Selected folder already contains log files, these will be overwritten
                     if you keep this selection"""
            tk.messagebox.showwarning("Warning", msg)
            window.destroy()

    def wind_forecast_changed(self, *args):
        """Callback function to handle wind forecast input being changed. Sets the max duration of
        the simulation to the duration of the forecast.
        """

        filepath = self.wind_forecast.get()

        if os.path.exists(filepath):
            with open(filepath, "rb") as f:
                wind_data = json.load(f)

            # get wind duration and set max_duration
            wind_time_step_min = wind_data['time_step_min']
            wind_time_step_sec = wind_time_step_min * 60
            num_entries = len(wind_data['data'])

            wind_duration_sec = wind_time_step_sec * num_entries

            time_unit = self.time_unit.get()

            if time_unit == "minutes":
                self.max_duration = wind_duration_sec / 60

            elif time_unit == "hours":
                self.max_duration = wind_duration_sec / 60 / 60

            self.duration_spinbox.configure(to=self.max_duration)

        elif filepath == "":
            pass

        else:
            self.wind_forecast.set("")
            window = tk.Tk()
            window.withdraw()
            tk.messagebox.showwarning("Error", "Selected folder does not contain a valid wind forecast!")
            window.destroy()

    def time_unit_changed(self, *args):
        """Callback function that handles the time unit selection being changed, adjusts the 
        displayed duration based on the unit
        """
        time_units = {"seconds": 1.0, "minutes": 60.0, "hours": 3600.0}

        # Get the ratio between the new and old time unit
        ratio = time_units[self.prev_t_unit] / time_units[self.time_unit.get()]

        self.max_duration *= ratio
        self.duration_spinbox.configure(to=self.max_duration)

        # Update sim_time
        current_sim_time = float(self.sim_time.get())
        new_sim_time = current_sim_time * ratio
        self.sim_time.set(new_sim_time)

        # Update prev_t_unit
        self.prev_t_unit = self.time_unit.get()

    def user_path_changed(self, *args):
        """Callback function that handles the user module path selection being changed, it gets
        a list of the classes within the module so that the class option menu can be refreshed.
        """

        user_module_file = self.user_path.get()
        user_module_dir = os.path.dirname(user_module_file)
        user_module_name = os.path.splitext(os.path.basename(user_module_file))[0]

        # Add the directory of the module to sys.path
        sys.path.insert(0, user_module_dir)

        try:
            # Dynamically import the module
            user_module = importlib.import_module(user_module_name)
            self.user_module = user_module

            self.class_options = [name for name, obj in inspect.getmembers(user_module)
                                if inspect.isclass(obj)
                                and obj.__module__ == user_module.__name__]

            self.update_option_menu()

        finally:
            # Remove the directory from sys.path
            sys.path.pop(0)

    def update_option_menu(self):
        """Function that changes the class option menu based on the module that is currently
        selected.
        """

        self.user_class_name.set(self.class_options[0])
        # Clear the current menu
        menu = self.class_opt_menu["menu"]
        menu.delete(0, "end")

        # Add the new options
        for class_option in self.class_options:
            menu.add_command(label=class_option,
                            command=lambda value=class_option: self.user_class_name.set(value))

    def class_changed(self, *args):
        """Callback function that handles the user class selection being changed, it checks to
        ensure the class selected is a valid subclass of ControlClass.
        """

        # Access the class within the module
        UserCodeClass = getattr(self.user_module, self.user_class_name.get())

        # Check if the user's class is a subclass of the abstract base class
        if not issubclass(UserCodeClass, ControlClass):
            menu = self.class_opt_menu["menu"]
            menu.delete(0, "end")
            window = tk.Tk()
            window.withdraw()
            tk.messagebox.showwarning("Warning", "User class must be instance of ControlClass!")
            window.destroy()

    def zero_wind_toggled(self, *args):
        if self.zero_wind.get():
            self.wind_button.configure(state='disabled')
            self.wind_entry.configure(state='disabled')
            self.wind_forecast.set("")

        else:
            self.wind_button.configure(state='normal')
            self.wind_entry.configure(state='normal')

    def validate_fields(self, *args):
        """Function used to validate the inputs, primarily responsible for activating/disabling
        the submit button based on if all necessary input has been provided.
        """
        if all([self.map_folder.get(), self.log_folder.get()]):
            self.submit_button.config(state='normal')
        else:
            self.submit_button.config(state='disabled')

    def submit(self):
        """Callback when the submit button is pressed. Stores all the relevant data in the result
        variable so it can be retrieved
        """
        sim_time_raw = self.sim_time.get()

        if self.time_unit.get() == "seconds":
            sim_time = sim_time_raw

        elif self.time_unit.get() == "minutes":
            sim_time = sim_time_raw * 60

        elif self.time_unit.get() == "hours":
            sim_time = sim_time_raw * 3600

        self.result = {
            "input": self.map_folder.get(),
            "log": self.log_folder.get(),
            "wind": self.wind_forecast.get(),
            "t_step": self.time_step.get(),
            "cell_size": self.cell_size.get(),
            "sim_time": sim_time,
            "viz_on": self.viz_on.get(),
            "num_runs": self.num_runs.get(),
            "user_path": self.user_path.get(),
            "class_name": self.user_class_name.get(),
            "zero_wind": self.zero_wind.get()
        }

        self.submit_callback(self.result)

class VizFolderSelector(FileSelectBase):
    """Class used to prompt user for log files to be visualized

    :param submit_callback: Function that should be called when the submit button is pressed
    :type submit_callback: Callable
    """
    def __init__(self, submit_callback: Callable):
        """Constructor method, populates tk window with all necessary elements and initializes
        necessary variables
        """
        super().__init__("Visualization Tool")

        # Define variables
        self.viz_folder = tk.StringVar()
        self.viz_folder.trace("w", self.viz_folder_changed)
        self.run_folder = tk.StringVar()
        self.run_folder.trace("w", self.run_folder_changed)
        self.viz_freq = tk.IntVar()
        self.viz_freq.set(60)
        self.scale_km = tk.DoubleVar()
        self.scale_km.set(1.0)
        self.legend = tk.BooleanVar()
        self.legend.set(True)
        self.has_agents = False
        self.run_folders = ["No runs available, select a folder"]

        self.init_location = False

        # Save submit callback function
        self.submit_callback = submit_callback

        frame = self.create_frame(self.root)

        # # Create a frame to select log file
        self.create_folder_selector(frame, "Log folder:    ", self.viz_folder)

        run_selection_frame = tk.Frame(frame)
        run_selection_frame.pack(padx=10,pady=5)
        tk.Label(run_selection_frame, text="Run to Visualize:",
                 anchor="center").grid(row=0, column=0)

        self.run_options = tk.OptionMenu(run_selection_frame, self.run_folder,
                      *self.run_folders)

        self.run_options.grid(row=0, column=1)

        # Create a frame to select sim time per frame
        self.create_spinbox_with_two_labels(frame, "Update Frequency:", np.inf,
                                            self.viz_freq, "seconds")

        # Create a frame to select scale bar size
        self.create_spinbox_with_two_labels(frame, "Scale bar size:        ", 100, self.scale_km, "km")

        # Create a frame to choose legend display
        legend_frame = tk.Frame(frame)
        legend_frame.pack(padx=10, pady=5)
        tk.Checkbutton(legend_frame, text="Display fuel legend",
                       variable=self.legend).grid(row=0,column=0)

        self.submit_button = tk.Button(frame, text='Submit', command = self.submit,
                                       state='disabled')
        self.submit_button.pack(pady=10)

    def viz_folder_changed(self, *args):
        """Callback function for selecting the log file to be displayed. Checks the file selected
        to make sure that all the required file types are there to visualize the data.
        # """
        folderpath = self.viz_folder.get()
        self.run_folders = self.get_run_sub_folders(folderpath)

        if self.run_folders:
            menu = self.run_options["menu"]
            menu.delete(0, "end")
            for run in self.run_folders:
                menu.add_command(label=run, command=lambda value=run: self.run_folder.set(value))

            self.run_folder.set(self.run_folders[0])

    def run_folder_changed(self, *args):
        """Callback function to handle the run folder input being changed. Ensures the selection
        contains valid log files.
        """
        run_foldername = self.run_folder.get()
        run_folderpath = os.path.join(self.viz_folder.get(), run_foldername)

        if os.path.exists(os.path.join(run_folderpath, 'agents.msgpack')):
            self.has_agents = True
            self.agent_file = os.path.join(run_folderpath, 'agents.msgpack')

        else:
            self.has_agents = False

        if os.path.exists(os.path.join(run_folderpath, 'log.msgpack')):
            self.viz_file = os.path.join(run_folderpath, 'log.msgpack')

        else:
            window = tk.Tk()
            window.withdraw()
            tk.messagebox.showwarning("Error",
            "Selected folder does not have a valid log file!")

            window.destroy()

            self.run_folder.set("")

    def get_run_sub_folders(self, folderpath: str) -> list:
        """Function that retrieves all the runs contained within a log folder. Returns a list of
        the folders to populate the option menu

        :param folderpath: string representing the path to the selected folder
        :type folderpath: str
        :return: list of run folder names inside the log folder with valid log files
        :rtype: list
        """

        if not os.path.exists(folderpath):
            return []

        if not os.path.exists(os.path.join(folderpath, 'init_fire_state.pkl')):
            window = tk.Tk()
            window.withdraw()
            tk.messagebox.showwarning("Error",
            "Selected file not in a folder with an initial state file!")

            window.destroy()

            return []

        self.init_location = True

        contents = os.listdir(folderpath)
        subfolders = [d for d in contents if os.path.isdir(os.path.join(folderpath, d)) and d.startswith("run")]

        return subfolders

    def submit(self):
        """Callback when the submit button is pressed. Stores all the relevant data in the result
        variable so it can be retrieved
        """
        self.result = {
            "file": self.viz_file,
            "freq": self.viz_freq.get(),
            "scale_km": self.scale_km.get(),
            "legend": self.legend.get(),
            "init_location": self.init_location,
            "has_agents": self.has_agents
        }

        if self.has_agents:
            self.result["agent_file"] = self.agent_file

        self.submit_callback(self.result)

    def validate_fields(self, *args):
        """Function used to validate the inputs, primarily responsible for activating/disabling
        the submit button based on if all necessary input has been provided.
        """
        if self.viz_folder.get() and self.viz_freq.get() > 0:
            self.submit_button.config(state='normal')
        else:
            self.submit_button.config(state='disabled')

class WindForecastGen(FileSelectBase):
    """Class used to allow users to generate wind forecasts

    :param submit_callback: Function that should be called when the submit button is pressed
    :type submit_callback: Callable
    """
    def __init__(self, submit_callback: Callable):
        """Constructor method, populates tk window with all necessary elements and initializes
        necessary variables
        """
        super().__init__("Wind Forecast Generator")

        self.submit_callback = submit_callback

        # define variables
        self.file_name = tk.StringVar()
        self.folderpath = tk.StringVar()
        self.duration = tk.DoubleVar()
        self.duration.set(1.0)
        self.time_step = tk.DoubleVar()
        self.curr_speed_val = tk.DoubleVar()
        self.curr_direction_val = tk.DoubleVar()

        self.wind_forecast = []

        self.time_step_last_changed = 0
        self.duration_last_changed = 0

        frame = self.create_frame(self.root)

        filename_frame = tk.Frame(frame)
        filename_frame.pack(padx=10,pady=5)
        tk.Label(filename_frame, text="Filename: ", anchor="center").grid(row=0, column=0)
        self.filename_entry = tk.Entry(filename_frame, textvariable=self.file_name, width = 25)
        self.filename_entry.grid(row=0, column=1)
        tk.Label(filename_frame, text=".json ", anchor="center").grid(row=0, column=2)

        self.create_folder_selector(frame, "Save forecast to: ", self.folderpath)

        entry_frame = tk.Frame(frame)
        entry_frame.pack(padx=10,pady=5)
        tk.Label(entry_frame, text="Wind Speed: ", anchor="center").grid(row=0, column=0)
        self.speed_entry = tk.Entry(entry_frame, textvariable=self.curr_speed_val, width = 10)
        self.speed_entry.grid(row=0, column=1)
        tk.Label(entry_frame, text="(m/s)", anchor="center").grid(row=0, column=2)

        time_step_frame = tk.Frame(frame)
        time_step_frame.pack(padx=20,pady=5)
        tk.Label(time_step_frame, text="Time Step: ", anchor="center").grid(row=0, column=0)
        self.t_step_entry = tk.Spinbox(time_step_frame, from_=1, to=1000, increment=1, textvariable=self.time_step, width=10)
        self.t_step_entry.grid(row=0, column=1)
        tk.Label(time_step_frame, text="mins", anchor="center").grid(row=0, column=2)

        # spacer
        tk.Label(time_step_frame, text="").grid(row=0, column=3, padx=20)

        tk.Label(time_step_frame, text="Duration: ", anchor="center").grid(row=0, column=4)
        self.duration_entry = tk.Spinbox(time_step_frame, from_=0.0166666, to=1000, increment=0.5, textvariable=self.duration, width=10)
        self.duration_entry.grid(row=0, column=5)
        tk.Label(time_step_frame, text="hours", anchor="center").grid(row=0, column=6)


        # spacer
        tk.Label(entry_frame, text="").grid(row=0, column=3, padx=20)

        tk.Label(entry_frame, text="Wind Direction: ", anchor="center").grid(row=0, column=4)
        self.dir_entry = tk.Entry(entry_frame, textvariable=self.curr_direction_val, width = 10)
        self.dir_entry.grid(row=0, column=5)
        tk.Label(entry_frame, text="degrees", anchor="center").grid(row=0, column=6)

        # Buttons to add, delete, and edit entries
        btn_add = tk.Button(entry_frame, text="Add", command=self.add_entry)
        btn_add.grid(row=0,column=7, padx=10)

        # Listbox to display entries
        self.listbox = tk.Listbox(frame, width=40, height=10)
        self.listbox.pack(pady=10)

        edit_del_frame = tk.Frame(frame)
        edit_del_frame.pack(padx=10,pady=5)

        btn_edit = tk.Button(edit_del_frame, text="Overwrite", command=self.overwrite_entry)
        btn_edit.grid(row=0, column=0)
        btn_delete = tk.Button(edit_del_frame, text="Delete", command=self.delete_entry)
        btn_delete.grid(row=0, column=1)

        # Create a submit button
        self.submit_button = tk.Button(frame, text="Save to File", command=self.submit, state='disabled')
        self.submit_button.pack(pady=10)

        self.time_step.trace('w', self.time_step_changed)
        self.duration.trace('w', self.duration_changed)

    def validate_fields(self):
        """Function used to validate the inputs, primarily responsible for activating/disabling
        the submit button based on if all necessary input has been provided.
        """
        folder_exists = os.path.exists(self.folderpath.get())
        filename_entered = self.file_name.get()
        entries_exist = len(self.wind_forecast)

        if entries_exist and self.time_step.get() > 0 and self.duration.get() > 0 and folder_exists and filename_entered:
            self.submit_button.config(state='normal')

    def add_entry(self):
        """Callback function for 'add' button that adds the current input for direction and speed
        to the working wind forecast.
        """

        entry = (self.curr_speed_val.get(), self.curr_direction_val.get())
        if entry:  # only add non-empty entries
            self.wind_forecast.append(entry)
            hours = int(np.floor(len(self.listbox.get(0, tk.END)) * self.time_step.get() / 60))
            mins = int(len(self.listbox.get(0, tk.END)) * self.time_step.get() % 60)
            formatted_entry = f"{hours} h {mins} m: {entry[0]} m/s, {entry[1]} deg"
            self.listbox.insert(tk.END, formatted_entry)

        self.duration.set(len(self.wind_forecast) * self.time_step.get() / 60)

        self.validate_fields()

    def delete_entry(self):
        """Callback function for 'delete' button that removes the selected entry from the working
        wind forecast.
        """
        try:
            # get the index of the selected entry
            index = self.listbox.curselection()[0]
            self.listbox.delete(index)

            del self.wind_forecast[index]

            self.duration.set(len(self.wind_forecast) * self.time_step.get() / 60)

        except IndexError:  # if no entry is selected
            pass

        self.validate_fields()

    def overwrite_entry(self):
        """Callback function for 'overwrite' button that overwrites the selected entry in the
        working wind forecast with the current input for speed and direction.
        """
        try:
            # get the index of the selected entry
            index = self.listbox.curselection()[0]

            updated_entry = (self.curr_speed_val.get(), self.curr_direction_val.get())

            self.wind_forecast[index] = updated_entry

            mins = index * self.time_step.get()

            hours = int(np.floor(mins / 60))
            mins = int(mins % 60)

            formatted_entry = f"{hours} h {mins} m: {updated_entry[0]} m/s, {updated_entry[1]} deg"

            self.listbox.delete(index)
            self.listbox.insert(index, formatted_entry)

        except IndexError:  # if no entry is selected
            pass

        self.validate_fields()

    def time_step_changed(self, *args):
        """Callback function that handles when the time step of the forecast has been changed.
        Ensures the entered time step is valid and updates the duration based on the new time 
        step and the current number of entries in the forecast.
        """
        try:
            current_time_step = self.time_step.get()
        except tk.TclError:
            current_time_step = 0  # default value if Spinbox is empty

        if current_time_step < 0:
            self.time_step.set(1)
            return

        self.listbox.delete(0, tk.END)
        for idx, (speed, direction) in enumerate(self.wind_forecast):
            mins = int((idx * current_time_step) % 60)
            hours = int(np.floor(idx * current_time_step) / 60)
            formatted_entry = f"{hours} h {mins} m: {speed} m/s, {direction} deg"
            self.listbox.insert(tk.END, formatted_entry)

        self.duration.set(len(self.wind_forecast) * current_time_step / 60)  # Convert to hours

        self.validate_fields()

    def duration_changed(self, *args):
        """Callback function that handles when the duration of the forecast has been changed.
        Ensures the entered duration is valid and updates the time step based on the new
        duration and the current number of entries in the forecast.
        """
        try:
            curr_duration = self.duration.get()
        except tk.TclError:
            curr_duration = 0  # default value if Spinbox is empty

        # If duration was the last changed variable, update time_step
        if self.duration_last_changed > self.time_step_last_changed:
            if len(self.wind_forecast) > 0:
                new_time_step = curr_duration * 60 / len(self.wind_forecast)  # Convert duration to minutes
                self.time_step.set(new_time_step)
        else:
            # Update duration based on new time_step and number of entries
            self.duration.set(len(self.wind_forecast) * self.time_step.get() / 60)  # Convert to hours

        # Update the last change time for duration
        self.duration_last_changed = time.time()

        self.validate_fields()

    def submit(self):
        """Callback when the submit button is pressed. Stores all the relevant data in the result
        variable so it can be retrieved
        """
        data = {
            "save_location": os.path.join(self.folderpath.get(), self.file_name.get() + ".json"),
            "time_step": self.time_step.get(),
            "forecast": self.wind_forecast
        }

        self.submit_callback(data)

        self.root.withdraw()
        self.root.quit()

class LoaderWindow:
    """Class used to created loading bars for progress updates to user while programs are running
    in backend

    :param title: title to be displayed at the top of the window
    :type title: str
    :param max_value: number of increments to complete the task
    :type max_value: int
    """
    def __init__(self, title: str, max_value: int):
        """Constructor method, populates tk window with all necessary elements and initializes
        necessary variables
        """
        self.root = tk.Toplevel()
        self.root.geometry('300x100')
        self.root.title(title)

        self.text_var = tk.StringVar()
        self.text_label = tk.Label(self.root, textvariable=self.text_var)
        self.text_label.pack()

        self.progressbar = ttk.Progressbar(self.root, maximum=max_value)
        self.progressbar.pack(fill='x')
        sleep(0.5)

    def set_text(self, text:str):
        """Sets the text of the window

        :param text: text to be displayed along the progress bar to inform user what task is being
                     completed
        :type text: str
        """
        sleep(0.5)
        self.text_var.set(text)
        self.root.update_idletasks()  # Ensure the window updates

    def increment_progress(self, increment_value:int=1):
        """Increments the progress forward by 'increment_value'

        :param increment_value: amount to increment forward, defaults to 1
        :type increment_value: int, optional
        """
        self.progressbar.step(increment_value)
        self.root.update_idletasks()  # Ensure the window updates

    def close(self):
        """Close the loader window
        """
        self.root.destroy()



"""Module responsible for visualization of simulations in real-time

.. autoclass:: Visualizer
    :members:
"""

import copy
from shapely.geometry import LineString
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.collections import PatchCollection
import matplotlib.patches as mpatches
from matplotlib.axes import Axes
from matplotlib import cm
import numpy as np

from embrs.utilities.fire_util import CellStates, FireTypes
from embrs.utilities.fire_util import FuelConstants as fc
from embrs.utilities.fire_util import RoadConstants as rc
from embrs.utilities.fire_util import UtilFuncs as util
from embrs.fire_simulator.fire import FireSim

mpl.use('TkAgg')

class Visualizer:
    """Class that visualizes simulations in real-time

    :param sim: :class:`~fire_simulator.fire.FireSim` instance to visualize
    :type sim: FireSim
    :param artists: list of artists that should be drawn initially, useful for quickly starting
                    numerous visualizations one after another, defaults to None
    :type artists: list, optional
    :param collections: list of collection objects that should be drawn initially, useful for
                        quickly starting numerous visualizations one after another, defaults to None
    :type collections: list, optional
    :param saved_legend: saved legend box, useful for quickly starting
                         numerous visualizations one after another, defaults to None
    :type saved_legend: Axes.legend, optional
    :param scale_bar_km: Determines how much distance the scale bar should represent in km,
                         defaults to 1.0
    :type scale_bar_km: int, optional
        """
    def __init__(self, sim: FireSim, artists:list=None, collections:list=None,
                 saved_legend:Axes.legend= None, scale_bar_km:float = 1.0):
        """Constructor method that initializes a visualization by populating all patches for cells
        and drawing all initial artists"""
        self.sim = sim

        width_m = sim.cell_size * np.sqrt(3) * sim.grid_width
        height_m = sim.cell_size * 1.5 * sim.grid_height

        self.agents = None
        self.agent_labels = None

        plt.ion()
        h_fig = plt.figure(figsize=(10, 10))
        h_ax = h_fig.add_axes([0.05, 0.05, 0.9, 0.9])

        # Create meshgrid for plotting contours
        x = np.arange(0, sim.shape[1])
        y = np.arange(0, sim.shape[0])
        X, Y = np.meshgrid(x, y)

        cont = h_ax.contour(X*sim.cell_size*np.sqrt(3),Y*sim.cell_size*1.5,
                            sim.coarse_topography, colors='k')

        h_ax.clabel(cont, inline=True, fontsize=10, zorder=2)

        if artists is None or collections is None:
            print("Initializing visualization... ")
            burnt_patches = []
            alpha_arr = [0, 1]
            break_fuel_arr = [0, 1]

            # Add low and high polygons to prevent weird color mapping
            r = 1/np.sqrt(3)
            low_poly = mpatches.RegularPolygon((-10,-10), numVertices=6, radius=r, orientation=0)
            high_poly = mpatches.RegularPolygon((-10,-10), numVertices=6, radius=r, orientation=0)
            fire_patches = [low_poly, high_poly]
            prescribe_patches = [low_poly, high_poly]
            tree_patches = [low_poly, high_poly]
            fire_breaks = [low_poly, high_poly]

            legend_elements = []
            added_colors = []

            # Add patches for each cell
            for i in range(sim.shape[0]):
                for j in range(sim.shape[1]):
                    curr_cell = sim.cell_grid[i][j]

                    polygon = mpatches.RegularPolygon((curr_cell.x_pos, curr_cell.y_pos),
                                     numVertices=6, radius=sim.cell_size, orientation=0)

                    if curr_cell.state == CellStates.FUEL:
                        color = fc.fuel_color_mapping[curr_cell.fuel_type.fuel_type]
                        if color not in added_colors:
                            added_colors.append(color)
                            legend_elements.append(mpatches.Patch(color = color,
                                                label = curr_cell.fuel_type.name))

                        if curr_cell.fuel_content < 1 and curr_cell.fuel_type.fuel_type < 90:
                            fire_breaks.append(polygon)
                            break_fuel_arr.append(curr_cell.fuel_content)

                        else:
                            polygon.set(color = color)
                            tree_patches.append(polygon)

                    elif curr_cell.state == CellStates.FIRE:
                        if curr_cell.fire_type == FireTypes.WILD:
                            fire_patches.append(polygon)
                            alpha_arr.append(curr_cell.fuel_content)
                        else:
                            prescribe_patches.append(polygon)

                    else:
                        burnt_patches.append(polygon)

            # Create collections grouping cells in each of their states
            tree_coll =  PatchCollection(tree_patches, match_original = True)

            if len(fire_breaks) > 0:
                breaks_coll = PatchCollection(fire_breaks, edgecolor='none')
                breaks_coll.set(array= break_fuel_arr, cmap=mpl.colormaps["gist_gray"])

            fire_coll = PatchCollection(fire_patches, edgecolor='none', facecolor='#F97306')
            if len(alpha_arr) > 0:
                alpha_arr = [float(i)/sum(alpha_arr) for i in alpha_arr]
                fire_coll.set(array=alpha_arr, cmap=mpl.colormaps["gist_heat"])

            prescribe_coll = PatchCollection(prescribe_patches, edgecolor='none', facecolor='b')

            burnt_coll = PatchCollection(burnt_patches, edgecolor='none', facecolor='k')

            self.collections = [copy.copy(breaks_coll), copy.copy(tree_coll),
                                copy.copy(fire_coll), copy.copy(burnt_coll),
                                copy.copy(prescribe_coll)]

            # Add collections to plot
            h_ax.add_collection(breaks_coll)
            h_ax.add_collection(tree_coll)
            h_ax.add_collection(fire_coll)
            h_ax.add_collection(burnt_coll)
            h_ax.add_collection(prescribe_coll)

            # Create time display
            time_box_x = 0
            time_box_y = sim.grid_height*1.5*sim.cell_size-(15/600) * height_m
            time_box_w = (1/6)*width_m
            time_box_h = (15/600)*height_m

            self.time_box = mpatches.Rectangle((time_box_x, time_box_y), time_box_w, time_box_h,
                                                facecolor='white', edgecolor='black', linewidth=1,
                                                zorder=3, alpha = 0.75)

            # Create wind display
            wind_box_x = 0
            wind_box_y = sim.grid_height*1.5*sim.cell_size - (650/6000) * height_m
            wind_box_w = (5/60)*width_m
            wind_box_h = (500/6000) * height_m

            self.wind_box = mpatches.Rectangle((wind_box_x, wind_box_y), wind_box_w, wind_box_h,
                                                facecolor='white', edgecolor ='black', linewidth=1,
                                                zorder = 3, alpha = 0.75)

            # Create scale display
            self.scale_box = mpatches.Rectangle((0, 10), 1100, (2/60)*height_m, facecolor='white',
                                                edgecolor='k', linewidth= 1, alpha=0.75, zorder= 3)
            
            # Add display items to artists
            self.artists = [copy.copy(self.time_box),
                            copy.copy(self.wind_box),
                            copy.copy(self.scale_box)]

            # Add display items to plot
            h_ax.add_patch(self.scale_box)
            h_ax.add_patch(self.wind_box)
            h_ax.add_patch(self.time_box)

            # Plot roads if they exist
            if sim.roads is not None:
                for road in sim.roads:
                    x_trimmed = []
                    y_trimmed = []

                    x, y = zip(*road[0])
                    for i, x_i in enumerate(x):
                        if 0 < x_i/30 < sim.grid_width and 0 < y[i]/30 < sim.grid_height:
                            x_trimmed.append(x_i)
                            y_trimmed.append(y[i])

                    x = tuple(xi for xi in x_trimmed)
                    y = tuple(yi for yi in y_trimmed)

                    road_color = rc.road_color_mapping[road[1]]
                    h_ax.plot(x, y, color= road_color)

                    if road_color not in added_colors:
                        added_colors.append(road_color)
                        legend_elements.append(mpatches.Patch(color=road_color,
                                               label = f"Road - {road[1]}"))

            # Plot firebreaks if they exist
            if sim.fire_breaks is not None:
                # Create a colormap for grey shades
                cmap = mpl.colormaps["Greys_r"]

                for fire_break in sim.fire_breaks:
                    line = fire_break['geometry']
                    fuel_val = fire_break['fuel_value']
                    if isinstance(line, LineString):
                        # Normalize the fuel_val between 0 and 1
                        normalized_fuel_val = fuel_val / 100.0
                        color = cmap(normalized_fuel_val)
                        x, y = line.xy
                        h_ax.plot(x, y, color=color)

            h_ax.legend(handles=legend_elements, loc='upper right', borderaxespad=0)
            self.legend_elements = legend_elements

        # Reload visualizer from initial state
        else:
            for coll in collections:
                coll = copy.copy(coll)
                h_ax.add_collection(coll)

            for artist in artists:
                artist = copy.copy(artist)
                h_ax.add_patch(artist)

            # Plot roads if they exist
            if sim.roads is not None:
                for road in sim.roads:
                    x_trimmed = []
                    y_trimmed = []

                    x, y = zip(*road[0])
                    for i, x_i in enumerate(x):
                        if 0 < x_i/30 < sim.grid_width and 0 < y[i]/30 < sim.grid_height:
                            x_trimmed.append(x_i)
                            y_trimmed.append(y[i])

                    x = tuple(xi for xi in x_trimmed)
                    y = tuple(yi for yi in y_trimmed)

                    road_color = fc.fuel_color_mapping[91]
                    h_ax.plot(x, y, color= road_color)

            h_ax.legend(handles=saved_legend, loc='upper right', borderaxespad=0)

        wx, wy = self.wind_box.get_xy()
        cx = wx + self.wind_box.get_width()/2

        self.windheader = h_ax.text(cx, wy + 0.85 * self.wind_box.get_height(),
                                    'Wind:', ha = 'center', va = 'center')

        time_str = util.get_time_str(sim.curr_time_s)

        rx, ry = self.time_box.get_xy()
        cx = rx + self.time_box.get_width()/2
        cy = ry + self.time_box.get_height()/2

        self.timeheader = h_ax.text(20, cy, 'time:', ha='left', va='center')

        self.simtext = h_ax.text(2*cx - 20, cy, time_str, ha='right', va='center')

        h_ax.set_aspect('equal')
        h_ax.axis([0, sim.cell_size*sim.shape[1]*np.sqrt(3) - (sim.cell_size*np.sqrt(3)/2),
                   0, sim.cell_size*1.5*sim.shape[0] - (sim.cell_size*1.5)])

        plt.tick_params(left = False, right = False, bottom = False,
                        labelleft = False, labelbottom = False)

        num_cells_scale = scale_bar_km * 1000

        if scale_bar_km < 1:
            scale_size = str(num_cells_scale) + "m"
        else:
            scale_size = str(scale_bar_km) + "km"

        scalebar = AnchoredSizeBar(h_ax.transData, num_cells_scale, scale_size, 'lower left',
                                   color ='k', pad = 0.1, frameon=False)
        h_ax.add_artist(scalebar)

        self.arrow_obj = None
        self.windtext = None

        self.h_ax = h_ax
        self.fig = h_fig

        self.fig.canvas.draw()
        self.initial_state = self.fig.canvas.copy_from_bbox(self.h_ax.bbox)

        plt.pause(1)

    def update_grid(self, sim: FireSim):
        """Updates the grid based on the current state of the simulation, this function is called
        at a frequency of sim.display_freq_s set in the FireSim constructor.

        :param sim: FireSim instance to display
        :type sim: FireSim
        """
        self.simtext.set_visible(False)

        fire_patches = []
        tree_patches = []
        burnt_patches = []
        alpha_arr = [0, 1]

        # Add low and high polygons to prevent weird color mapping
        r = 1/np.sqrt(3)
        low_poly = mpatches.RegularPolygon((-10, -10), numVertices=6, radius=r,orientation=0)
        high_poly = mpatches.RegularPolygon((-10, -10), numVertices=6, radius=r,orientation=0)
        fire_patches = [low_poly, high_poly]
        tree_patches = [low_poly, high_poly]
        prescribe_patches = [low_poly, high_poly]

        soak_xs = []
        soak_ys = []
        c_vals = []

        for c in sim.updated_cells.values():
            polygon = mpatches.RegularPolygon((c.x_pos, c.y_pos), numVertices=6,
                                              radius=sim.cell_size, orientation=0)

            if c.state == CellStates.FUEL:
                color = np.array(list(mcolors.to_rgba(fc.fuel_color_mapping[c.fuel_type.fuel_type])))
                # Scale color based on cell's fuel content
                color = color *  c.fuel_content
                polygon.set_facecolor(color)
                tree_patches.append(polygon)

                if c.dead_m > 0.08: # fuel moisture not nominal
                    soak_xs.append(c.x_pos)
                    soak_ys.append(c.y_pos)
                    c_val = c.dead_m/fc.dead_fuel_moisture_ext_table[c.fuel_type.fuel_type]
                    c_val = np.min([1, c_val])
                    c_vals.append(c_val)

            elif c.state == CellStates.FIRE:
                if c.fire_type == FireTypes.WILD:
                    fire_patches.append(polygon)
                    alpha_arr.append(c.fuel_content)
                else:
                    prescribe_patches.append(polygon)

            else:
                burnt_patches.append(polygon)

        color_map = cm.get_cmap('Blues')
        norm = mcolors.Normalize(vmin=0, vmax=1)
        self.h_ax.scatter(soak_xs, soak_ys, c=c_vals, cmap = color_map, marker='2', norm=norm)

        tree_patches = np.array(tree_patches)
        fire_patches = np.array(fire_patches)
        prescribe_patches = np.array(prescribe_patches)
        burnt_patches = np.array(burnt_patches)
        alpha_arr = np.array(alpha_arr)

        tree_coll =  PatchCollection(tree_patches, match_original=True)

        fire_coll = PatchCollection(fire_patches, edgecolor='none', facecolor='#F97306')
        if len(alpha_arr) > 0:
            fire_coll.set(array=alpha_arr, cmap=mpl.colormaps["gist_heat"])

        prescribe_coll = PatchCollection(prescribe_patches, edgecolor='none', facecolor='pink')

        burnt_coll = PatchCollection(burnt_patches, edgecolor='none', facecolor='k')

        self.h_ax.add_collection(tree_coll)
        self.h_ax.add_collection(fire_coll)
        self.h_ax.add_collection(burnt_coll)
        self.h_ax.add_collection(prescribe_coll)

        if self.arrow_obj is not None:
            self.arrow_obj.remove()

        if self.windtext is not None:
            self.windtext.remove()

        wx, wy = self.wind_box.get_xy()
        cx = wx + self.wind_box.get_width()/2
        cy = wy + self.wind_box.get_height()/2

        self.windtext = self.h_ax.text(cx, wy + 0.1 * self.wind_box.get_height(),
                                       str(np.round(sim.wind_vec.wind_speed,2)) + " m/s",
                                       ha = 'center', va = 'center')

        if sim.wind_vec.wind_speed != 0:
            wind_dir_vec = sim.wind_vec.vec/np.linalg.norm(sim.wind_vec.vec)
            dx = wind_dir_vec[0]
            dy = wind_dir_vec[1]
            arrow_len = self.wind_box.get_height()/3

            self.arrow_obj = self.h_ax.arrow(cx-(arrow_len*dx) , cy-(arrow_len*dy),
                                             dx*arrow_len, dy*arrow_len, width=10,
                                             head_width = 50, color = 'r', zorder= 3)
        else:
            self.arrow_obj = self.h_ax.text(cx, cy, 'X', fontsize = 20, color = 'r',
                                            ha='center', va = 'center')

        self.h_ax.draw_artist(self.windtext)
        self.h_ax.draw_artist(self.arrow_obj)

        sim_time_s = sim.time_step*sim.iters
        time_str = util.get_time_str(sim_time_s)

        rx, ry = self.time_box.get_xy()
        cx = rx + self.time_box.get_width()/2
        cy = ry + self.time_box.get_height()/2

        self.simtext = self.h_ax.text(2*cx - 20, cy, time_str, ha='right', va='center')

        # Plot agents at current time if they exist
        if len(sim.agent_list) > 0:
            if self.agents is not None:
                for a in self.agents:
                    a.remove()
            if self.agent_labels is not None:
                for label in self.agent_labels:
                    label.remove()

            self.agents = []
            for agent in sim.agent_list:
                a = self.h_ax.scatter(agent.x, agent.y, marker=agent.marker, color=agent.color)
                self.agents.append(a)

            self.agent_labels = []
            for agent in sim.agent_list:
                if agent.label is not None:
                    label = self.h_ax.annotate(agent.label, (agent.x, agent.y))
                    self.agent_labels.append(label)

        sim.updated_cells.clear()

        self.fig.canvas.blit(self.h_ax.bbox)
        self.fig.canvas.flush_events()

    def reset_figure(self, done=False):
        """Resets the figure back to its initial state. Used to reset the figure between different
        simulation runs

        :param done: Closes the figure when set to True, resets and continues to next visualization
                     if False, defaults to False
        :type done: bool, optional
        """
        # Close figure
        plt.close(self.fig)

        if not done:
            self.__init__(self.sim, self.artists,
                          self.collections, self.legend_elements)

    def visualize_prediction(self, prediction_grid):
        print("visualzing prediction...")

        time_steps = sorted(prediction_grid.keys())

        cmap = mpl.colormaps["viridis"]
        norm = plt.Normalize(time_steps[0], time_steps[-1])

        for time in prediction_grid.keys():
            x,y = zip(*prediction_grid[time])
            self.h_ax.scatter(x,y, c=[cmap(norm(time))])

        print("Finished visualizing....")



"""Module that handles user drawing on top of sim map when specifying map parameters.
"""

from typing import Tuple
import numpy as np
from PyQt5.QtWidgets import QApplication, QInputDialog
from matplotlib.widgets import Button
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('TkAgg')

class PolygonDrawer:
    """Class used for drawing polygons on top of sim map for specifying locations of initial
    ignitions and fire-breaks.

    :param fig: matplotlib figure object used to draw on top of
    :type fig: matplotlib.figure.Figure
    """
    def __init__(self, fig: matplotlib.figure.Figure):
        """Constructor method that initializes all variables and sets up the GUI
        """
        self.fig = fig

        if fig.axes:
            self.ax = fig.axes[0]  # Get the existing Axes object if available
        else:
            self.ax = fig.subplots()

        self.num_road_lines = len(self.ax.lines)

        self.xlims = self.ax.get_xlim()
        self.ylims = self.ax.get_ylim()

        self.ax.set_title("Click on the map to draw a polygon to specify the initial fire region")

        self.ax.invert_yaxis()
        self.line, = self.ax.plot([], [], 'r-')  # Create a line for confirmed segments
        self.preview_line, = self.ax.plot([], [], 'r--')  # Create a line for the preview segment
        self.xs = []
        self.ys = []
        self.polygons = []  # List of polygons
        self.temp_polygon = None  # For storing the temporary polygon
        self.decision_pending = False

        # Create accept/decline buttons but keep them invisible until a polygon is closed
        self.accept_button = Button(plt.axes([0.78, 0.05, 0.1, 0.075]), 'Accept')
        self.accept_button.on_clicked(self.accept)
        self.accept_button.ax.set_visible(False)

        self.decline_button = Button(plt.axes([0.885, 0.05, 0.1, 0.075]), 'Decline')
        self.decline_button.on_clicked(self.decline)
        self.decline_button.ax.set_visible(False)

        # Create apply/clear buttons but keep them inactive until a polygon or line is accepted
        self.apply_button = Button(plt.axes([0.785, 0.95, 0.1, 0.04]), 'Apply')
        self.apply_button.on_clicked(self.apply)
        self.apply_button.set_active(False)
        self.apply_button.color = '0.85'
        self.apply_button.hovercolor = self.apply_button.color

        self.clear_button = Button(plt.axes([0.89, 0.95, 0.1, 0.04]), 'Clear')
        self.clear_button.on_clicked(self.clear)
        self.clear_button.set_active(False)
        self.clear_button.color = '0.85'
        self.clear_button.hovercolor = self.clear_button.color

        # Create no fire breaks button but keep it invisible until polygons are specified
        self.no_fire_breaks_button = Button(plt.axes([0.78, 0.05, 0.1, 0.075]), 'No Fire Breaks')
        self.no_fire_breaks_button.on_clicked(self.skip_fire_breaks)
        self.no_fire_breaks_button.ax.set_visible(False)

        # Create reset view button that resets view to original
        self.reset_view_button = Button(plt.axes([0.12, 0.05, 0.1, 0.075]), 'Reset View')
        self.reset_view_button.on_clicked(self.reset_view)

        # Set up event handlers
        self.cid_press = self.fig.canvas.mpl_connect('button_press_event', self.on_press)
        self.cid_motion = self.fig.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.cid_scroll = self.fig.canvas.mpl_connect('scroll_event', self.on_scroll)
        self.cid_release = self.fig.canvas.mpl_connect('button_release_event', self.on_release)

        # Additional initialization for line segments
        self.line_segments = []  # List of line segments
        self.current_line = None  # For storing the current line segment
        self.temp_line_segments = []  # For storing temporary line segments
        self.lines = []
        self.fire_break_fuel_vals = []

        # Parameter to track whether in ignition or fire-break mode
        self.mode = 'ignition'

        self.valid = False

    def on_press(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback for handling button presses in the figure, left click places either initial
        ignitions or fire-breaks, right click can be used to pan the view.

        :param event: MouseEvent triggered from the click
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if event.inaxes in [self.accept_button.ax,
                            self.decline_button.ax,
                            self.apply_button.ax,
                            self.clear_button.ax,
                            self.reset_view_button.ax]:
            return

        # Left click allows user to draw on map
        if event.button == 1:
            if self.decision_pending:
                return

            if event.xdata is None or event.ydata is None:
                return

            if self.mode == 'ignition':
                # If there are already points, check if the new point closes a polygon
                if len(self.xs) > 2:
                    dx = self.xs[0] - event.xdata
                    dy = self.ys[0] - event.ydata
                    # If the point is close to the first point, close the polygon
                    if np.hypot(dx, dy) < 1:
                        self.ax.set_title("Press 'Accept' to confirm or 'Decline' to discard")

                        # Store polygon vertices
                        self.temp_polygon = list(zip(self.xs, self.ys))

                        # Append the first point to the lists of points
                        self.xs.append(self.xs[0])
                        self.ys.append(self.ys[0])

                        # Fill the polygon
                        self.ax.fill(self.xs, self.ys, 'r', alpha=0.5)
                        self.line.set_data(self.xs, self.ys)

                        # Clear the x an y values, and preview line
                        self.xs = []
                        self.ys = []
                        self.preview_line.set_data([], [])

                        # Make buttons visible once polygon is closed
                        self.accept_button.ax.set_visible(True)
                        self.decline_button.ax.set_visible(True)
                        self.decision_pending = True
                        self.fig.canvas.draw()

                        return

                self.xs.append(event.xdata)
                self.ys.append(event.ydata)
                self.line.set_data(self.xs, self.ys)

                self.set_button_status(False, False)

                self.fig.canvas.draw()

            elif self.mode == 'fire-breaks':
                self.no_fire_breaks_button.ax.set_visible(False)

                if len(self.xs) > 0:
                    self.temp_line_segments.append([self.xs[-1], self.ys[-1]])
                    self.temp_line_segments.append([event.xdata, event.ydata])
                self.xs.append(event.xdata)
                self.ys.append(event.ydata)
                self.line.set_data(self.xs, self.ys)

                if len(self.xs) > 1:
                    self.accept_button.ax.set_visible(True)
                    self.decline_button.ax.set_visible(True)
                    title = """Click 'Accept' to confirm or 'Decline' to discard. Or continue
                               to draw fire-break"""

                    self.ax.set_title(title)
                else:
                    self.accept_button.ax.set_visible(False)
                    self.decline_button.ax.set_visible(False)

                self.set_button_status(False, False)

                self.fig.canvas.draw()

        # Right click allows user to pan the view
        elif event.button == 3:
            self.ax._pan_start = [event.xdata, event.ydata]

    def on_release(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling mouse release, only used when panning the view

        :param event: MouseEvent triggered by releasing the mouse
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if event.button == 3:
            self.ax._pan_start = None

    def on_motion(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling mouse motion. Pans the view when right-click active,
        previews lines to be drawn otherwise

        :param event: MouseEvent triggered by moving mouse
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if event.inaxes in [self.accept_button.ax,
                            self.decline_button.ax,
                            self.apply_button.ax,
                            self.clear_button.ax]:
            return

        if event.button == 3 and self.ax._pan_start is not None:
            dx = event.xdata - self.ax._pan_start[0]
            dy = event.ydata - self.ax._pan_start[1]
            xlim = self.ax.get_xlim()
            ylim = self.ax.get_ylim()

            self.ax.set_xlim(xlim[0] - dx, xlim[1] - dx)
            self.ax.set_ylim(ylim[0] - dy, ylim[1] - dy)

            self.fig.canvas.draw()

        else:
            if self.decision_pending:
                return

            if event.xdata is None or event.ydata is None:
                return
            if not self.xs:  # If no points have been confirmed, don't draw a preview
                return

            if self.mode == 'ignition':
                dx = self.xs[0] - event.xdata
                dy = self.ys[0] - event.ydata

                # If the mouse is close to the first point, snap to the first point
                if np.hypot(dx, dy) < 1:
                    preview_xs = [self.xs[-1], self.xs[0]]
                    preview_ys = [self.ys[-1], self.ys[0]]

                # If the mouse is not close to the first point, draw to the current mouse position
                else:
                    preview_xs = [self.xs[-1], event.xdata]
                    preview_ys = [self.ys[-1], event.ydata]

                self.preview_line.set_data(preview_xs, preview_ys)
                self.fig.canvas.draw()

            elif self.mode == 'fire-breaks':
                # If no points have been confirmed, don't draw a preview
                if not self.xs:
                    return
                preview_xs = [self.xs[-1], event.xdata]
                preview_ys = [self.ys[-1], event.ydata]
                self.preview_line.set_data(preview_xs, preview_ys)

            self.fig.canvas.draw()

    def on_scroll(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling scrolling in the figure. Zooms the view in and out
        focused wherever the mouse pointer is.

        :param event: MouseEvent triggered by scrolling
        :type event: matplotlib.backend_bases.MouseEvent
        """
        cur_xlim = self.ax.get_xlim()
        cur_ylim = self.ax.get_ylim()

        xdata = event.xdata  # get event x location
        ydata = event.ydata  # get event y location

        if xdata is None or ydata is None:
            return

        xleft = xdata - cur_xlim[0]
        xright = cur_xlim[1] - xdata
        ybottom = ydata - cur_ylim[0]
        ytop = cur_ylim[1] - ydata

        if event.button == 'up':
            scale_factor = 1/1.1
        elif event.button == 'down':
            scale_factor = 1.1
        else:
            scale_factor = 1

        self.ax.set_xlim([xdata - xleft*scale_factor,
                    xdata + xright*scale_factor])
        self.ax.set_ylim([ydata - ybottom*scale_factor,
                    ydata + ytop*scale_factor])

        self.fig.canvas.draw()

    def set_button_status(self, apply_active: bool, clear_active: bool):
        """Set the status of the 'apply' and 'clear' buttons to set whether they are active or not

        :param apply_active: boolean to set 'apply' button status, True = active, False = inactive
        :type apply_active: bool
        :param clear_active: boolean to set 'clear' button status, True = active, False = inactive
        :type clear_active: bool
        """
        self.apply_button.set_active(apply_active)
        self.apply_button.color = '0.75' if apply_active else '0.85'
        self.apply_button.hovercolor = '0.95' if apply_active else '0.85'
        self.apply_button.ax.set_facecolor(self.apply_button.color)

        self.clear_button.set_active(clear_active)
        self.clear_button.color = '0.75' if clear_active else '0.85'
        self.clear_button.hovercolor = '0.95' if clear_active else '0.85'
        self.clear_button.ax.set_facecolor(self.clear_button.color)

    def reset_current_polygon(self):
        """Remove all drawn polygons
        """
        self.xs = []
        self.ys = []
        for patch in self.ax.patches:
            patch.remove()

    def reset_current_lines(self):
        """Remove all drawn lines
        """
        self.xs = []
        self.ys = []
        for line in self.ax.lines[self.num_road_lines:]:
            if line not in (self.line, self.preview_line):
                line.remove()

        self.line.set_data([], [])
        self.preview_line.set_data([],[])

    def decline(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling the decline button being pressed. Clears the most recent
        polygon or line drawn

        :param event: MouseEvent triggered from clicking 'decline'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if self.mode == 'ignition':
            self.xs = []
            self.ys = []
            self.ax.patches.pop()
            self.hide_buttons()
            title = "Click on the map to draw a polygon to specify the initial fire region"
            self.ax.set_title(title)
            self.line.set_data([], [])
            self.preview_line.set_data([], [])
            self.decision_pending = False

            if self.polygons:
                self.set_button_status(True, True)

            self.fig.canvas.draw()
        elif self.mode == 'fire-breaks':
            self.temp_line_segments = []
            self.current_line = None
            for line in self.lines:
                line.remove()
            self.lines = []
            self.preview_line.set_data([], [])
            self.line.set_data([], [])
            self.xs = []
            self.ys = []

            self.accept_button.ax.set_visible(False)
            self.decline_button.ax.set_visible(False)

            if len(self.line_segments) == 0:
                self.no_fire_breaks_button.ax.set_visible(True)

            self.fig.canvas.draw()

    def clear(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling the clear button being pressed. Clears all polygons or
        all lines depending on the current mode

        :param event: MouseEvent triggered from clicking 'clear'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if self.mode == 'ignition':
            self.polygons = []
            self.reset_current_polygon()
            title = "Click on the map to draw a polygon to specify the initial fire region"
            self.ax.set_title(title)

        elif self.mode == 'fire-breaks':
            self.line_segments = []
            self.fire_break_fuel_vals = []
            self.reset_current_lines()
            self.ax.set_title("Draw line segments to specify fire-breaks")
            self.no_fire_breaks_button.ax.set_visible(True)

        self.set_button_status(False, False)
        self.fig.canvas.draw()

    def accept(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling the accept button being pressed. Confirms the most recent
        polygon or line.

        :param event: MouseEvent triggered from clicking 'accept'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if self.mode == 'ignition':
            self.polygons.append(self.temp_polygon)
            self.temp_polygon = None
            self.hide_buttons()
            self.ax.plot()

            if self.polygons:
                self.set_button_status(True, True)

            title = "Draw another fire region or click 'Apply' to save changes and move on"
            self.ax.set_title(title)

        elif self.mode == 'fire-breaks':
            if self.temp_line_segments:
                self.line_segments.append(self.temp_line_segments)
                self.temp_line_segments = []
                self.set_button_status(True, True)

                val = self.get_fuel_value()

                self.fire_break_fuel_vals.append(val)

            self.hide_buttons()
            title = "Draw another fire-break line or click 'Apply' to save changes and finish"
            self.ax.set_title(title)
            self.ax.plot(self.xs, self.ys, 'b')  # plot the line segment

        self.xs = []
        self.ys = []
        self.line.set_data([], [])
        self.preview_line.set_data([], [])
        self.decision_pending = False

        self.fig.canvas.draw()

    def reset_view(self, event: matplotlib.backend_bases.MouseEvent):
        """Resets the view to the original display

        :param event: MouseEvent triggered from pressing 'reset_view'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        self.ax.set_xlim(self.xlims)
        self.ax.set_ylim(self.ylims)
        self.ax.invert_yaxis()

        self.fig.canvas.draw()

    def get_fuel_value(self) -> float:
        """Prompts user for the fuel value of a just drawn fire-break

        :return: float fuel value entered by the user
        :rtype: float
        """
        app = QApplication([])
        request  = "Enter percent fuel remaining in fire break:"
        value, ok = QInputDialog.getDouble(None, "Input Dialog", request)
        if ok:
            return value

        return None

    def apply(self, event: matplotlib.backend_bases.MouseEvent):
        """Callback function for handling the apply button being pressed. Saves the polygons or
        lines drawn in permanent data structures, closes the figure if process is complete

        :param event: MouseEvent triggered from clicking 'apply'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        if self.mode == 'ignition':
            self.mode = 'fire-breaks'
            self.ax.set_title("Draw line segments to specify fire-breaks")
            self.no_fire_breaks_button.ax.set_visible(True)
            self.fig.canvas.draw()
            self.preview_line, = self.ax.plot([], [], 'b--')
            self.line, = self.ax.plot([], [], 'b-')

        elif self.mode == 'fire-breaks':
            self.valid = True
            plt.close(self.fig)

    def skip_fire_breaks(self, event: matplotlib.backend_bases.MouseEvent):
        """Skips the drawing of the fire-breaks and closes the figure

        :param event: MouseEvent triggered from clicking 'skip fire breaks'
        :type event: matplotlib.backend_bases.MouseEvent
        """
        self.line_segments = []
        self.fire_break_fuel_vals = []
        self.valid = True
        plt.close(self.fig)

    def hide_buttons(self):
        """Hide the accept and decline buttons from view
        """
        self.accept_button.ax.set_visible(False)
        self.decline_button.ax.set_visible(False)
        self.fig.canvas.draw()

    def get_ignitions(self) -> list:
        """Get the ignition polygons that have been finalized

        :return: list of polygon coordinates representing ignition areas
        :rtype: list
        """

        return self.polygons

    def get_fire_breaks(self) -> Tuple[list, list]:
        """Get the fire breaks drawn and finalized along with their fuel values

        :return: Returns a list with the coordinates of the fire-break line segments, and a list
                 corresponding to each of their fuel values
        :rtype: Tuple[list, list]
        """

        return self.line_segments, self.fire_break_fuel_vals



"""Sample class showing how a custom agent class can be created using the AgentBase class.

There is nothing to run in this example, to see how it is used see the next two examples 'vi_controlled_burning.py
and 'vii_suppression.py'
"""

import numpy as np

from base_classes.agent_base import AgentBase
from utilities.fire_util import FuelConstants as fc

class SampleAgent(AgentBase):
    def __init__(self, id: any, x: float, y: float, fire, label=None, marker='*', color='magenta'):
        super().__init__(id, x, y, label, marker=marker, color=color)

        # Define useful variables for agent class
        self.fire = fire
        self.task_list = []
        self.curr_task = None
        self.t_step = fire.time_step # seconds
        self.speed = 10 # m/s
        self.max_tasks = 0
        self.tasks_done = False

    def complete_tasks(self):
        if len(self.task_list) > 0:
            if self.curr_task is not None:

                # get current x and y
                x1 = self.x
                y1 = self.y

                # get x and y of next point
                x2 = self.curr_task[0][0]
                y2 = self.curr_task[0][1]

                # calculate total distance
                dx_total = x2 - x1
                dy_total = y2 - y1
                d_total = np.linalg.norm((dx_total, dy_total))

                # calculate incremental vector pointing at next point
                dir = np.arctan2(y2 - y1, x2 - x1)
                lin = self.t_step * self.speed

                # get dx and dy
                dx = lin * np.cos(dir)
                dy = lin * np.sin(dir)

                # check if dx, dy will overshoot the point
                if lin >= d_total:
                    # overshot so set x and y to goal position
                    self.x = x2
                    self.y = y2

                    if self.curr_task[1] == 'burn':
                        # Carry out burnout operation
                        self.fire.set_prescribed_fire_at_xy(self.x, self.y)
                        self.tasks_done = True

                    elif self.curr_task[1] == 'soak':
                        # Carry out suppression operation
                        cell = self.fire.get_cell_from_xy(self.x, self.y)
                        max_moisture = fc.dead_fuel_moisture_ext_table[cell.fuel_type.fuel_type]
                        self.fire.set_fuel_moisture_at_xy(self.x, self.y, max_moisture)
                        self.tasks_done = True

                    self.task_list.remove(self.curr_task)
                    self.curr_task = None

                else:
                    # Update agent position
                    self.x += dx
                    self.y += dy

            else:
                # Find next task based on which is closest
                min_dist = np.inf
                best_task = None
                for task in self.task_list:

                    dist = np.sqrt((self.x - task[0][0])**2 + (self.y - task[0][1])**2)

                    if dist < min_dist:
                        best_task = task

                self.curr_task = best_task


"""Various sets of constants and helper functions useful throughout the codebase

.. autoclass:: UtilFuncs
    :members:

.. autoclass:: CellStates
    :members:

.. autoclass:: FireTypes
    :members:

.. autoclass:: FuelConstants
    :members:

.. autoclass:: RoadConstants
    :members:

.. autoclass:: HexGridMath
    :members:

.. autoclass:: ControlledBurnParams
    :members:

.. autoclass:: WindAdjustments
    :members:
"""

from typing import Tuple
import numpy as np
from shapely.geometry import Polygon, MultiPolygon
from shapely.ops import unary_union

class CellStates:
    """Enumeration of the possible cell states.

    Attributes:
        - **BURNT** (int): Represents a cell that has been burnt and has no fuel remaining.
        - **FUEL** (int): Represents a cell that still contains fuel and is not on fire.
        - **FIRE** (int): Represents a cell that is currently on fire.
    """
    # Cell States:
    BURNT, FUEL, FIRE = 0, 1, 2

class FireTypes:
    """Enumeration of the possible fire types.

    Attributes:
        - **WILD** (int): Represents a fire that occurred naturally.
        - **PRESCRIBED** (int): Represents a fire that was started in a controlled manner, burns with lower intensity.

    """
    # Fire Types:
    WILD, PRESCRIBED = 0, 1

class FuelConstants:
    """Various values and dictionaries pertaining to modelling of fuel types.

    Attributes:
        - **ch_h_to_m_min** (float): float that converts chains/hr to m/min.
        - **fbfm_13_keys** (list): list of ints corresponding to each of Anderson's 13 FBFMs.
        - **fuel_names** (dict): dictionary where keys are ints for each FBFM and values are the names of each fuel type.
        - **fuel_type_revers_lookup** (dict): dictionary where keys are the fuel type names and values are the ints for each FBFM.
        - **init_fuel_table** (dict): dictionary of the initial fuel content of each fuel type.
        - **nom_spread_prob_table** (dict): dictionary of dictionaries for each fuel type's nominal spread probabilities.
        - **nom_spread_vel_table** (dict): dictionary of nominal spread velocities for wildfires for each fuel type.
        - **dead_fuel_moisture_ext_table** (dict): dictionary of each fuel type's dead fuel moisture of extinction.
        - **fuel_consumption_factor_table** (dict): dictionary of each fuel type's mass-loss curve weighting factor.
        - **fuel_color_mapping** (dict): dictionary mapping each fuel type to the display color for visualizations.

    """
    # Conversion from chains/hr to m/min
    ch_h_to_m_min = 0.33528

    # Conversion from inches to mm
    in_to_mm = 25.4

    # Valid keys for the FBFM 13 fuel model
    fbfm_13_keys = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 91, 92, 93, 98, 99]

    # Dictionary of fuel number to name
    fuel_names = {1: "Short grass", 2: "Timber grass", 3: "Tall grass", 4: "Chaparral",
                5: "Brush", 6: "Hardwood slash", 7: "Southern rough", 8: "Closed timber litter",
                9: "Hardwood litter", 10: "Timber litter", 11: "Light logging slash",
                12: "Medium logging slash", 13: "Heavy logging slash", 91: 'Urban', 92: 'Snow/ice',
                93: 'Agriculture', 98: 'Water', 99: 'Barren'}

    fuel_type_reverse_lookup = {"Short grass": 1, "Timber grass": 2, "Tall grass": 3, "Chaparral": 4,
                "Brush": 5, "Hardwood slash": 6, "Southern rough": 7 , "Closed timber litter": 8,
                "Hardwood litter": 9, "Timber litter": 10, "Light logging slash": 11,
                "Medium logging slash": 12, "Heavy logging slash": 13, 'Urban': 91, 'Snow/ice': 92,
                'Agriculture': 93, 'Water': 98, 'Barren': 99}

    # Standard initial fuel contents for each type
    init_fuel_table = {1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 9: 1, 10: 1, 11: 1,
                    12: 1, 13: 1, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0}

    # Table of nominal spread probabilities between each fuel type
    nom_spread_prob_table = {
        1: {1: 0.425, 2: 0.45, 3: 0.475, 4: 0.475, 5: 0.4375, 6: 0.4625, 7: 0.425, 8: 0.425,
            9: 0.475, 10: 0.425, 11: 0.45, 12: 0.45, 13: 0.475, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        2: {1: 0.45, 2: 0.475, 3: 0.5, 4: 0.5, 5: 0.4625, 6: 0.4875, 7: 0.45, 8: 0.45, 9: 0.5,
            10: 0.45, 11: 0.475, 12: 0.475, 13: 0.5, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        3: {1: 0.475, 2: 0.5, 3: 0.525, 4: 0.525, 5: 0.4875, 6: 0.5125, 7: 0.475, 8: 0.475,
            9: 0.525, 10: 0.475, 11: 0.5, 12: 0.5, 13: 0.525, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        4: {1: 0.35, 2: 0.375, 3: 0.4, 4: 0.425, 5: 0.3875, 6: 0.4125, 7: 0.375, 8: 0.375,
            9: 0.425, 10: 0.4, 11: 0.4, 12: 0.425, 13: 0.45, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        5: {1: 0.3125, 2: 0.3375, 3: 0.3625, 4: 0.3875, 5: 0.35, 6: 0.375, 7: 0.3375, 8: 0.3375,
            9: 0.3875, 10: 0.3625, 11: 0.3625, 12: 0.3875, 13: 0.4125, 91: 0, 92: 0, 93: 0, 98: 0,
            99: 0},

        6: {1: 0.3375, 2: 0.3625, 3: 0.3875, 4: 0.4125, 5: 0.375, 6: 0.4, 7: 0.3625, 8: 0.3625,
            9: 0.4125, 10: 0.3875, 11: 0.3875, 12: 0.4125, 13: 0.4375, 91: 0, 92: 0, 93: 0, 98: 0,
            99: 0},

        7: {1: 0.3, 2: 0.325, 3: 0.35, 4: 0.375, 5: 0.3375, 6: 0.3625, 7: 0.325, 8: 0.325,
            9: 0.375, 10: 0.35, 11: 0.35, 12: 0.375, 13: 0.4, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        8: {1: 0.025, 2: 0.05, 3: 0.075, 4: 0.1, 5: 0.0625, 6: 0.0875, 7: 0.05, 8: 0.025, 9: 0.075,
            10: 0.225, 11: 0.05, 12: 0.25, 13: 0.275, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        9: {1: 0.075, 2: 0.1, 3: 0.125, 4: 0.15, 5: 0.1125, 6: 0.1375, 7: 0.1, 8: 0.075, 9: 0.125,
            10: 0.275, 11: 0.1, 12: 0.3, 13: 0.325, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        10: {1: 0.2, 2: 0.225, 3: 0.25, 4: 0.325, 5: 0.2875, 6: 0.3125, 7: 0.275, 8: 0.3, 9: 0.35,
            10: 0.3, 11: 0.325, 12: 0.325, 13: 0.35, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        11: {1: 0.05, 2: 0.075, 3: 0.1, 4: 0.125, 5: 0.0875, 6: 0.1125, 7: 0.075, 8: 0.05, 9: 0.1,
            10: 0.25, 11: 0.075, 12: 0.275, 13: 0.3, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        12: {1: 0.225, 2: 0.25, 3: 0.275, 4: 0.35, 5: 0.3125, 6: 0.3375, 7: 0.3, 8: 0.325,
            9: 0.375, 10: 0.325, 11: 0.35, 12: 0.35, 13: 0.375, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        13: {1: 0.25, 2: 0.275, 3: 0.3, 4: 0.375, 5: 0.3375, 6: 0.3625, 7: 0.325, 8: 0.35, 9: 0.4,
            10: 0.35, 11: 0.375, 12: 0.375, 13: 0.4, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        91: {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0,
            10: 0, 11: 0, 12: 0, 13: 0, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        92: {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0,
            10: 0, 11: 0, 12: 0, 13: 0, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        93: {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0,
            10: 0, 11: 0, 12: 0, 13: 0, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        98: {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0,
            10: 0, 11: 0, 12: 0, 13: 0, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},

        99: {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0,
            10: 0, 11: 0, 12: 0, 13: 0, 91: 0, 92: 0, 93: 0, 98: 0, 99: 0},
    }

    # values based on 5 mi/h wind, given in chains/hr
    nom_spread_vel_table = {1: 78, 2: 35, 3: 104, 4: 75, 5: 18, 6: 32, 7: 20, 8: 1.6,
                                    9: 7.5, 10: 7.9, 11: 6, 12: 13, 13: 13.5, 91: 0, 92: 0,
                                    93: 0, 98: 0, 99: 0}

    # based on the values in the Anderson Fuel model
    dead_fuel_moisture_ext_table = {1: 0.12, 2: 0.15, 3: 0.25, 4: 0.20, 5: 0.20, 6: 0.25, 7: 0.40,
                                    8: 0.30, 9: 0.25, 10: 0.25, 11: 0.15, 12: 0.20, 13: 0.25,
                                    91: 0.1, 92: 0, 93: 0, 98: 0, 99: 0}

    # Weighting factor for mass-loss curve for each fuel type
    fuel_consumption_factor_table = {1: 7, 2: 30, 3: 7, 4: 360, 5: 360, 6: 360, 7: 360,
                            8: 1200, 9: 1200, 10: 1200, 11: 1200, 12: 1200, 13: 1200,
                            91: 0, 92: 0, 93: 0, 98: 0, 99: 0}

    # Color mapping for each fuel type
    fuel_color_mapping = {1: 'xkcd:pale green', 2:'xkcd:lime', 3: 'xkcd:bright green',
                        4: 'xkcd:teal', 5: 'xkcd:bluish green', 6: 'xkcd:greenish teal',
                        7: 'xkcd:light blue green', 8: 'xkcd:pale olive' , 9: 'xkcd:olive',
                        10: 'xkcd:light forest green', 11: 'xkcd:bright olive',
                        12: 'xkcd:tree green', 13: 'xkcd:avocado green', 91: 'xkcd:ugly purple',
                        92: 'xkcd:pale cyan' , 93: "xkcd:perrywinkle", 98: 'xkcd:water blue',
                        99: 'xkcd:black'}


class WindAdjustments:
    """Mapping of wind speed to Lorentzian curve parameters, used to calculate alpha_w

    Attributes:
        - **wind_speed_param_mapping** (dict): dictionary mapping wind speeds (km/h) to Lorentzian parameters (A, gamma, C)
    """

    wind_speed_param_mapping = {0: (0, 1, 1), 10: (0.33, 114.93, 0.89), 20: (1.17, 84.83, 0.64), 30: (2.12, 60.95, 0.4),
                                50: (2.8, 45.13, 0.25), 60: (3.03, 39.61, 0.2), 70: (3.26, 33.92, 0.15), 80: (3.49, 27.56, 0.1),
                                90: (3.72, 20, 0.05), 100: (3.94, 7.71, 0.01)}


class ControlledBurnParams:
    """Parameters differentiating :py:attr:`~FireTypes.PRESCRIBED` fires from :py:attr:`~FireTypes.WILD`

    Attributes:
        - **nominal_prob_adj** (float): factor adjusting the nominal spread probability.
        - **nominal_vel_adj** (float): factor adjusting the nominal spread velocity.
        - **consumption_factor_adj** (float): factor adjusting the fuel consumption factor W.
        - **min_burnable_fuel_content** (float): minimum fuel fraction remaining in a neighboring cell for a possible ignition
        - **burnout_fuel_frac** (float): fraction of fuel at ignition remaining in burning cell at which it stops burning.
    """
    nominal_prob_adj = 0.75
    nominal_vel_adj = 0.5
    consumption_factor_adj = 1.5
    min_burnable_fuel_content = 0.25
    burnout_fuel_frac = 0.3

class RoadConstants:
    """Information on the modelling of different road types.

    Attributes:
        - **major_road_types** (list): list of the names of each of the major road types defined by `OpenStreetMap <https://wiki.openstreetmap.org/wiki/United_States/Road_classification>`_.
        - **road_fuel_vals** (dict): dictionary mapping road types to the amount of fuel modelled along them.
        - **road_color_mapping** (dict): dictionary mapping road types to their display color for visualizations.

    """
    # Road types
    major_road_types = ['motorway', 'trunk' , 'primary', 'secondary',
                        'tertiary', 'unclassified', 'residential']

    # Fuel values for each road type
    road_fuel_vals = {'motorway': 0, 'trunk': 0 , 'primary': 0.1, 'secondary': 0.15,
                      'tertiary': 0.2, 'unclassified': 0.3, 'residential': 0.25}

    road_color_mapping = {
        'motorway': '#4B0082',  # Indigo
        'trunk': '#800080',    # Purple
        'primary': '#9400D3',  # DarkViolet
        'secondary': '#9932CC',  # DarkOrchid
        'tertiary': '#BA55D3',  # MediumOrchid
        'unclassified': '#DA70D6',  # Orchid
        'residential': '#EE82EE',  # Violet
    }

class HexGridMath:
    """Data structures to help with handling cell neighbors in a hexagonal grid.

    Attributes:
        - **hex_angle** (float): float representing the angle between neighboring hexagon cell centers.
        - **even_neighborhood_mapping** (dict): dictionary that maps relative indices of a cell's neighbors to unit vectors between them, for even rows.
        - **odd_neighborhood mapping** (dict): dictionary that maps relative indices of a cell's neighbors to unit vectors between them, for odd rows.
        - **even_neighborhood** (list): list of relative indices of a cell's neighbors, for even rows.
        - **odd_neighborhood** (list): list of relative indices of a cell's neighbors, for odd rows.


    """
    # Define angle between neighboring hex centers
    hex_angle = 60 * (np.pi/180)

    # Define neighborhoods and unit vector mappings for even and odd rows
    even_neighborhood_mapping = {(-1,1): (-np.cos(hex_angle), np.sin(hex_angle)),
                                (0, 1): (np.cos(hex_angle), np.sin(hex_angle)),
                                (1,0): (1,0), (0, -1): (np.cos(hex_angle), -np.sin(hex_angle)),
                                (-1, -1): (-np.cos(hex_angle), -np.sin(hex_angle)),
                                (-1,0): (-1,0)}

    odd_neighborhood_mapping = {(1,0): (1,0), (1,1): (np.cos(hex_angle), np.sin(hex_angle)),
                                (0,1): (-np.cos(hex_angle), np.sin(hex_angle)), (-1,0): (-1,0),
                                (0,-1): (-np.cos(hex_angle), -np.sin(hex_angle)),
                                (1, -1): (np.cos(hex_angle), -np.sin(hex_angle))}

    even_neighborhood = [(-1,1), (0, 1), (1,0), (0, -1), (-1, -1), (-1,0)]
    odd_neighborhood = [(1,0), (1,1), (0,1), (-1,0), (0,-1), (1, -1)]

class UtilFuncs:
    """Various utility functions that are useful across numerous files.
    """
    def get_indices_from_xy(x_m: float, y_m: float, cell_size: float, grid_width: int,
                            grid_height: int) -> Tuple[int, int]:
        """Get the row and column indices in a backing array of a cell containing the point
        (x_m, y_m). 
        
        Does not require a :class:`~fire_simulator.fire.FireSim` object, uses 'cell_size' and the size of the array
        to calculate indices.

        :param x_m: x position in meters where indices should be found
        :type x_m: float
        :param y_m: y position in meters where indices should be found
        :type y_m: float
        :param cell_size: cell size in meters, measured as the distance across two parallel sides
                          of a regular hexagon
        :type cell_size: float
        :param grid_width: number of columns in the backing array of interest
        :type grid_width: int
        :param grid_height: number of rows in the backing array of interest
        :type grid_height: int
        :raises ValueError: if x or y inputs are out of bounds for the array constructed by
                            'cell_size', 'grid_width', and 'grid_height'
        :return: tuple containing [row, col] indices at the point (x_m, y_m)
        :rtype: Tuple[int, int]
        """
        row = int(y_m // (cell_size * 1.5))

        if row % 2 == 0:
            col = int(x_m // (cell_size * np.sqrt(3))) + 1
        else:
            col = int((x_m // (cell_size * np.sqrt(3))) - 0.5) + 1

        if col < 0 or row < 0 or row >= grid_height or col >= grid_width:
            msg = (f'Point ({x_m}, {y_m}) is outside the grid. '
                f'Column: {col}, Row: {row}, '
                f'simSize: ({grid_height} , {grid_width})')
            raise ValueError(msg)

        return row, col

    def get_time_str(time_s: int, show_sec = False) -> str:
        """Returns a formatted time string in m-h-s format from the time in seconds.
        
        Useful for generating readable display of the time.

        :param time_s: time value in seconds
        :type time_s: int
        :param show_sec: set to `True` if seconds should be displayed, `False` if not, defaults to `False`
        :type show_sec: bool, optional
        :return: formatted time string in h-m-s format
        :rtype: str
        """
        hours = int(time_s // 3600)
        minutes = int((time_s % 3600) // 60)

        if show_sec:
            seconds = int((time_s % 3600) % 60)

            if hours > 0:
                result = f"{hours} h {minutes} min {seconds} s"
            elif minutes > 0:
                result = f"{minutes} min {seconds} s"
            else:
                result = f"{seconds} s"
            return result

        if hours > 0:
            result = f"{hours} h {minutes} min"
        else:
            result = f"{minutes} min"
        return result

    def get_dominant_fuel_type(fuel_map: np.ndarray) -> int:
        """Finds the most commonly occurring fuel type within a fuel map.

        :param fuel_map: Fuel map for a region
        :type fuel_map: np.ndarray
        :return: Integer representation of the dominant fuel type. 
            See :py:attr:`~utilities.fire_util.FuelConstants.fuel_names`
        :rtype: int
        """
        counts = np.bincount(fuel_map.ravel())

        return np.argmax(counts)

    def get_cell_polygons(cells: list) -> list:
        """Converts a list of cell objects into the minimum number of :py:attr:`shapely.Polygon`
        required to describe all of them

        :param cells: list of :class:`~fire_simulator.cell.Cell` objects to be converted
        :type cells: list
        :return: list of :py:attr:`shapely.Polygon` representing the cells
        :rtype: list
        """

        if not cells:
            return None

        polygons = [Polygon(UtilFuncs.hexagon_vertices(cell.x_pos, cell.y_pos, cell.cell_size)) for cell in cells]

        merged_polygon = unary_union(polygons)

        if isinstance(merged_polygon, MultiPolygon):
            return list(merged_polygon.geoms)

        return [merged_polygon]

    @staticmethod
    def hexagon_vertices(x: float, y: float, s: float) -> list:
        """Calculates the locations of each of a hexagons vertices with center (x,y) and side
        length s.

        :param x: x location of hexagon center
        :type x: float
        :param y: y location of hexagon center
        :type y: float
        :param s: length of hexagon sides
        :type s: float
        :return: list of (x,y) points representing the hexagon's vertices
        :rtype: list
        """
        vertices = [
            (x, y + s),
            (x + s * np.sqrt(3) / 2, y + s / 2),
            (x + s * np.sqrt(3) / 2, y - s / 2),
            (x, y - s),
            (x - s * np.sqrt(3) / 2, y - s / 2),
            (x - s * np.sqrt(3) / 2, y + s / 2)
        ]
        return vertices



"""Sample class which carries out prescribed burns based on the current wind conditions
and the locations of the fire breaks and roads.

Note: This example is not intended to be a robust firefighting algorithm, it is to be used
as a reference on how agents can be used in conjuction with the fire interface to formulate
a response to a fire. It does not, for example, handle the case where there is no wind.

This example works well when using the provided example map titled 'burnout_map' and the sample
wind forecast titled 'burnout_wind_forecast'

To run this example code, start a fire sim and select this file as the "User Module"
"""

import heapq
import numpy as np

from base_classes.control_base import ControlClass
from fire_simulator.fire import FireSim
from v_sample_agent import SampleAgent

class BurnoutExample(ControlClass):
	def __init__(self, fire: FireSim):
		# Define time agents should dispatch
		self.operation_start_time_h = 0.33

		# Initialize empty list to store agents
		self.agents = []

		# Define number of agents
		num_agents = 10

		# Randomly place agents
		for i in range(num_agents):
			a_x = np.random.random() * fire.x_lim
			a_y = np.random.random() * fire.y_lim

			a = SampleAgent(i, a_x, a_y, fire, label=f"{i}")
			a.max_tasks = 10

			fire.add_agent(a)
			self.agents.append(a)

		self.burnout_complete = False

	# process_state is called by the FireSim every iteration
	def process_state(self, fire: FireSim) -> None:
		# Check if it is time to dispatch agents
		if fire.curr_time_h > self.operation_start_time_h:
			if not self.burnout_complete:
				# Create tasks
				burn_tasks = self.create_burn_tasks(fire)
				
				# Assign tasks
				self.assign_burn_tasks(burn_tasks, 50, fire)

				self.burnout_complete = True

		# Step each agent forward one time step
		self.iterate_agents()

	def create_burn_tasks(self, fire: FireSim) -> list:
		# Get the average fire coordinate
		avg_fire_x, avg_fire_y = fire.get_avg_fire_coord()

		# Retrieve current wind conditions
		wind_vec = fire.get_curr_wind_vec()
		wind_mag = np.linalg.norm(wind_vec	)
		wind_unit_vec = wind_vec/wind_mag

		# Retrieve locations of fire breaks
		fire_break_points = []
		for cell in fire.fire_break_cells:	
			fire_break_points.append((cell.x_pos, cell.y_pos))

		# Retrieve locations of roads
		if fire.roads is not None:
			for road in fire.roads:
				for point in road[0]:
					road_x, road_y = point[0], point[1]
					fire_break_points.append((road_x, road_y))

		# Create vectors from avg fire position to the fire break and road points
		fire_break_vectors = [[p[0] - avg_fire_x, p[1] - avg_fire_y] for p in fire_break_points]

		# Calculate angles between fire break vectors and wind
		angles = [vector_angle(wind_unit_vec, fbv) for fbv in fire_break_vectors]
		
		# Filter out angles that are outside of 90 degrees from the wind direction
		within_cone = [fire_break_points[i] for i, angle in enumerate(angles) if angle <= 90]

		# Create vector that points in oppostive direction of wind
		projection_vec = -500*np.array(wind_unit_vec)

		# Project points from fire breaks using the above vector, these are used as task points
		task_points = [[p[0] + projection_vec[0], p[1] + projection_vec[1]] for p in within_cone]

		return task_points
	
	def assign_burn_tasks(self, tasks, spacing, fire):
		tasks_copy = tasks.copy()

		# List to keep track of all assigned tasks
		assigned_tasks = []  

		# Get average fire coordinate
		fire_x, fire_y = fire.get_avg_fire_coord()

		# Create empty list of task penalties
		tasks_penalties = np.empty(len(tasks_copy))

		# Calculate task penalties for each task based on distance from avg fire coordinate
		for i, t in enumerate(tasks_copy):
			t_dist = np.sqrt((t[0] - fire_x)**2 + (t[1] - fire_y)**2)
			tasks_penalties[i] = t_dist

		# Assign each agent with tasks
		for agent in self.agents:
			# Calculate distances from agent to all tasks
			distances = [np.sqrt((agent.x - task[0])**2 +
						(agent.y - task[1])**2) for task in tasks_copy]

			# Create a min-heap of distances and indices
			k = 2
			heap = [(dist+(k*tasks_penalties[i]), i) for i, dist in enumerate(distances)]
			heapq.heapify(heap)

			while len(agent.task_list) < agent.max_tasks and heap:
				# Get the task with the smallest distance
				_, best_task_index = heapq.heappop(heap)
				# Check if this task is > 50m away from all tasks already assigned to any agent
				best_task = tasks_copy[best_task_index]
				if all(np.sqrt((t[0] - best_task[0])**2 + (t[1] - best_task[1])**2) > spacing
						for t in assigned_tasks):

					# Add best task to the agent's task list
					agent.task_list.append((best_task, 'burn'))
					# Also add it to the list of all assigned tasks
					assigned_tasks.append(best_task)

	# Call the complete_tasks function for each agent each time-step
	def iterate_agents(self):
		for agent in self.agents:
			agent.complete_tasks()

# Function to calculate angle between vectors
def vector_angle(v1, v2):
    dot_product = np.dot(v1, v2)
    mags = np.linalg.norm(v1) * np.linalg.norm(v2)
    return np.degrees(np.arccos(dot_product/mags))



"""Demo class demonstrating the use of custom logging during a fire simulation.
These log messages will appear in the "messages" section of the status_log.json file

To run this example code, start a fire sim and select this file as the "User Module"
"""

from base_classes.control_base import ControlClass
from fire_simulator.fire import FireSim


class CustomLogging(ControlClass):
	def __init__(self, fire:FireSim):
		# Log custom messages with fire.logger.log_message()
		fire.logger.log_message("Custom Logging Code Running")

		self.operation_performed = False
		self.area_messaged_logged = False

	# process_state is called by the FireSim every iteration
	def process_state(self, fire: FireSim) -> None:
		# Log a message when some operation is performed
		if fire.curr_time_h > 1 and not self.operation_performed:
			self.perform_some_operation(fire)
		
		# Calculate burning area of the fire
		burning_area = len(fire.curr_fires) * fire.cell_grid[0, 0].cell_area

		# Log a message when burning area exceeds a certain value
		if burning_area > 5000 and not self.area_messaged_logged:
			fire.logger.log_message(f"Burning area exceeded 5000 m^2")
			self.area_messaged_logged = True
		
	def perform_some_operation(self, fire):
		fire.logger.log_message(f"Some operation performed at {fire.curr_time_h} hours")
		self.operation_performed = True


"""Representation of fuel types for the simulator

.. autoclass:: Fuel
    :members:
"""

from embrs.utilities.fire_util import FuelConstants as fc
from embrs.utilities.fire_util import CellStates

class Fuel:
    """Class representation of a fuel type, contains various relevant fuel properites to model
    fire spread

    :param fuel_type: numerical value of fuel type, corresponds to 13 Anderson FBFMs
    :type fuel_type: int
    :param init_state: initial state of the cell for given fuel type, defaults to
                       :py:attr:`CellStates.FUEL`
    :type init_state: :class:`~utilities.fire_util.CellStates`, optional
    """
    def __init__(self, fuel_type: int, init_state = CellStates.FUEL):
        """Constructor method to initialize a :class:`~fire_simulator.fuel.Fuel` instance.
        Calculates relevant properties from

        :class:`~utilities.fire_util.FuelConstants`.
        """
        self._fuel_type = fuel_type
        self._name = fc.fuel_names[fuel_type]

        # Grab all relevant fuel properties
        self._spread_probs = fc.nom_spread_prob_table[fuel_type]
        self._nominal_vel = fc.nom_spread_vel_table[fuel_type]*fc.ch_h_to_m_min
        self._consumption_factor = fc.fuel_consumption_factor_table[fuel_type]
        self._dead_m_ext = fc.dead_fuel_moisture_ext_table[fuel_type]

        self._init_state = init_state
        self._init_fuel = fc.init_fuel_table[fuel_type]

    def __str__(self) -> str:
        """To string method of :class:`fire_simulator.fuel.Fuel` class.

        :return: String representation of the fuel type.
        :rtype: str
        """
        return f"type: {self.name}"

    @property
    def fuel_type(self) -> int:
        """Numerical representation of fuel type, corresponds to 13 Anderson FBFMs.

        Possible values: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 91, 92, 93, 98, 99

        """
        return self._fuel_type

    @property
    def name(self) -> str:
        """String representing the name of the fuel type
        """
        return self._name

    @property
    def spread_probs(self) -> dict:
        """Dictionary containing the nominal spread probability from the current fuel type
        to all others. 
        
        The keys of the dictionary are the integer values representing all the fuel types. 
        The values are the spread probability between the current instance's fuel type and
        all the others.
        """
        return self._spread_probs

    @property
    def nominal_vel(self) -> float:
        """Nominal spread velocity of a wildfire in the fuel type (m/min)
        """
        return self._nominal_vel

    @property
    def consumption_factor(self) -> float:
        """Weighting factor for mass-loss curve based on 
        `(Coen 2005) <http://dx.doi.org/10.1071/WF03043>`_
        """
        return self._consumption_factor

    @property
    def dead_m_ext(self) -> float:
        """The dead fuel moisture of extinction for the fuel type
        """
        return self._dead_m_ext

    @property
    def init_state(self) -> CellStates:
        """The initial state of the fuel type
        """
        return self._init_state

    @property
    def init_fuel(self) -> float:
        """The initial amount of fuel a fuel type contains (between 0 and 1)
        """
        return self._init_fuel



import matplotlib.pyplot as plt
import matplotlib.patches as patches
import textwrap

fuel_names = {
    1: "Short grass", 2: "Timber grass", 3: "Tall grass", 4: "Chaparral",
    5: "Brush", 6: "Hardwood slash", 7: "Southern rough", 8: "Closed timber litter",
    9: "Hardwood litter", 10: "Timber litter", 11: "Light logging slash",
    12: "Medium logging slash", 13: "Heavy logging slash", 91: 'Urban', 92: 'Snow/ice',
    93: 'Agriculture', 98: 'Water', 99: 'Barren'
}

fuel_color_mapping = {
    1: 'xkcd:pale green', 2:'xkcd:lime', 3: 'xkcd:bright green', 4: 'xkcd:teal',
    5: 'xkcd:bluish green', 6: 'xkcd:greenish teal', 7: 'xkcd:light blue green',
    8: 'xkcd:pale olive', 9: 'xkcd:olive', 10: 'xkcd:light forest green',
    11: 'xkcd:bright olive', 12: 'xkcd:tree green', 13: 'xkcd:avocado green',
    91: 'xkcd:ugly purple', 92: 'xkcd:pale cyan', 93: "xkcd:perrywinkle",
    98: 'xkcd:water blue', 99: 'xkcd:black'
}

def hexagon(x_center, y_center, size):
    """Generate the coordinates for a hexagon."""
    return [(x_center + size * sin(theta), y_center + size * cos(theta))
            for theta in [i * (2 * pi / 6) for i in range(6)]]


# Initialize the plot
fig, ax = plt.subplots()
ax.set_aspect('equal')
ax.axis('off')

from math import sin, cos, pi

radius = 1.0
vertical_spacing = 2.5
horizontal_spacing = 4 * radius
text_width = 12

# Starting coordinates
x, y = 0, 0 

for idx, key in enumerate(fuel_names):
    hex_coords = hexagon(x, y, radius)
    hex_patch = patches.Polygon(hex_coords, facecolor=fuel_color_mapping[key], edgecolor='black')
    ax.add_patch(hex_patch)
    
    # Add wrapped label
    wrapped_text = textwrap.fill(fuel_names[key], text_width)
    ax.text(x, y - 2 * radius, wrapped_text, ha='center', va='center', fontsize=8)

    # Update x, y for next hexagon
    if (idx + 1) % 6 == 0:  # Go to the next row after every 6 hexagons
        x = 0
        y -= vertical_spacing * 2  # double vertical spacing to move to the next row
    else:
        x += horizontal_spacing

ax.autoscale_view()
plt.show()